@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package awsauth
 
 import (
@@ -13,6 +16,8 @@ import (
 
 	"github.com/aws/aws-sdk-go/aws/session"
 	"github.com/aws/aws-sdk-go/service/sts"
+	"github.com/stretchr/testify/assert"
+
 	"github.com/hashicorp/vault/sdk/logical"
 )
 
@@ -121,9 +126,129 @@ func TestBackend_pathLogin_parseIamArn(t *testing.T) {
 	}
 }
 
-func TestBackend_validateVaultHeaderValue(t *testing.T) {
+func TestBackend_validateVaultGetRequestValues(t *testing.T) {
 	const canaryHeaderValue = "Vault-Server"
-	requestURL, err := url.Parse("https://sts.amazonaws.com/")
+
+	getHeadersMissing := http.Header{
+		"Host": []string{"Foo"},
+	}
+	getHeadersInvalid := http.Header{
+		"Host":            []string{"Foo"},
+		iamServerIdHeader: []string{"InvalidValue"},
+	}
+	getHeadersValid := http.Header{
+		"Host":            []string{"Foo"},
+		iamServerIdHeader: []string{canaryHeaderValue},
+	}
+	getQueryValid := url.Values(map[string][]string{
+		"X-Amz-Algorithm":  {"AWS4-HMAC-SHA256"},
+		"X-Amz-Credential": {"AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request"},
+		amzSignedHeaders:   {"host;x-vault-aws-iam-server-id"},
+		"X-Amz-Signature":  {"5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
+		"X-Amz-User-Agent": {"aws-sdk-go-v2/1.2.0 os/linux lang/go/1.16 md/GOOS/linux md/GOARCH/amd64"},
+		"Action":           {"GetCallerIdentity"},
+		"Version":          {"2011-06-15"},
+	})
+	getQueryUnsigned := url.Values(map[string][]string{
+		"X-Amz-Algorithm":  {"AWS4-HMAC-SHA256"},
+		"X-Amz-Credential": {"AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request"},
+		amzSignedHeaders:   {"host"},
+		"X-Amz-Signature":  {"5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
+		"X-Amz-User-Agent": {"aws-sdk-go-v2/1.2.0 os/linux lang/go/1.16 md/GOOS/linux md/GOARCH/amd64"},
+		"Action":           {"GetCallerIdentity"},
+		"Version":          {"2011-06-15"},
+	})
+	getQueryNoAction := url.Values(map[string][]string{
+		"X-Amz-Algorithm":  {"AWS4-HMAC-SHA256"},
+		"X-Amz-Credential": {"AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request"},
+		amzSignedHeaders:   {"host;x-vault-aws-iam-server-id"},
+		"X-Amz-Signature":  {"5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
+		"X-Amz-User-Agent": {"aws-sdk-go-v2/1.2.0 os/linux lang/go/1.16 md/GOOS/linux md/GOARCH/amd64"},
+		"Version":          {"2011-06-15"},
+	})
+	getQueryInvalidAction := url.Values(map[string][]string{
+		"X-Amz-Algorithm":  {"AWS4-HMAC-SHA256"},
+		"X-Amz-Credential": {"AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request"},
+		amzSignedHeaders:   {"host;x-vault-aws-iam-server-id"},
+		"X-Amz-Signature":  {"5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
+		"X-Amz-User-Agent": {"aws-sdk-go-v2/1.2.0 os/linux lang/go/1.16 md/GOOS/linux md/GOARCH/amd64"},
+		"Action":           {"GetSessionToken"},
+		"Version":          {"2011-06-15"},
+	})
+	getQueryMultipleActions := url.Values(map[string][]string{
+		"X-Amz-Algorithm":  {"AWS4-HMAC-SHA256"},
+		"X-Amz-Credential": {"AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request"},
+		amzSignedHeaders:   {"host;x-vault-aws-iam-server-id"},
+		"X-Amz-Signature":  {"5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
+		"X-Amz-User-Agent": {"aws-sdk-go-v2/1.2.0 os/linux lang/go/1.16 md/GOOS/linux md/GOARCH/amd64"},
+		"Action":           {"GetCallerIdentity;GetSessionToken"},
+		"Version":          {"2011-06-15"},
+	})
+	validGetRequestURL, err := url.Parse("https://sts.amazonaws.com/?" + getQueryValid.Encode())
+	if err != nil {
+		t.Fatalf("error parsing test URL: %v", err)
+	}
+	unsignedGetRequestURL, err := url.Parse("https://sts.amazonaws.com/?" + getQueryUnsigned.Encode())
+	if err != nil {
+		t.Fatalf("error parsing test URL: %v", err)
+	}
+	noActionGetRequestURL, err := url.Parse("https://sts.amazonaws.com/?" + getQueryNoAction.Encode())
+	if err != nil {
+		t.Fatalf("error parsing test URL: %v", err)
+	}
+	invalidActionGetRequestURL, err := url.Parse("https://sts.amazonaws.com/?" + getQueryInvalidAction.Encode())
+	if err != nil {
+		t.Fatalf("error parsing test URL: %v", err)
+	}
+	multipleActionsGetRequestURL, err := url.Parse("https://sts.amazonaws.com/?" + getQueryMultipleActions.Encode())
+	if err != nil {
+		t.Fatalf("error parsing test URL: %v", err)
+	}
+
+	err = validateVaultHeaderValue(http.MethodGet, getHeadersMissing, validGetRequestURL, canaryHeaderValue)
+	if err == nil {
+		t.Error("validated GET request with missing Vault header")
+	}
+
+	err = validateVaultHeaderValue(http.MethodGet, getHeadersInvalid, validGetRequestURL, canaryHeaderValue)
+	if err == nil {
+		t.Error("validated GET request with invalid Vault header value")
+	}
+
+	err = validateVaultHeaderValue(http.MethodGet, getHeadersValid, unsignedGetRequestURL, canaryHeaderValue)
+	if err == nil {
+		t.Error("validated GET request with unsigned Vault header")
+	}
+
+	err = validateLoginIamRequestUrl(http.MethodGet, noActionGetRequestURL)
+	if err == nil {
+		t.Error("validated GET request with no Action parameter")
+	}
+
+	err = validateLoginIamRequestUrl(http.MethodGet, multipleActionsGetRequestURL)
+	if err == nil {
+		t.Error("validated GET request with multiple Action parameters")
+	}
+
+	err = validateLoginIamRequestUrl(http.MethodGet, invalidActionGetRequestURL)
+	if err == nil {
+		t.Error("validated GET request with an invalid Action parameter")
+	}
+
+	err = validateLoginIamRequestUrl(http.MethodGet, validGetRequestURL)
+	if err != nil {
+		t.Errorf("did NOT validate valid GET request: %v", err)
+	}
+
+	err = validateVaultHeaderValue(http.MethodGet, getHeadersValid, validGetRequestURL, canaryHeaderValue)
+	if err != nil {
+		t.Errorf("did NOT validate valid GET request: %v", err)
+	}
+}
+
+func TestBackend_validateVaultPostRequestValues(t *testing.T) {
+	const canaryHeaderValue = "Vault-Server"
+	postRequestURL, err := url.Parse("https://sts.amazonaws.com/")
 	if err != nil {
 		t.Fatalf("error parsing test URL: %v", err)
 	}
@@ -146,39 +271,93 @@ func TestBackend_validateVaultHeaderValue(t *testing.T) {
 		iamServerIdHeader: []string{canaryHeaderValue},
 		"Authorization":   []string{"AWS4-HMAC-SHA256 Credential=AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request, SignedHeaders=content-type;host;x-amz-date;x-vault-aws-iam-server-id, Signature=5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
 	}
-
 	postHeadersSplit := http.Header{
 		"Host":            []string{"Foo"},
 		iamServerIdHeader: []string{canaryHeaderValue},
 		"Authorization":   []string{"AWS4-HMAC-SHA256 Credential=AKIDEXAMPLE/20150830/us-east-1/iam/aws4_request", "SignedHeaders=content-type;host;x-amz-date;x-vault-aws-iam-server-id, Signature=5d672d79c15b13162d9279b0855cfba6789a8edb4c82c400e06b5924a6f2b5d7"},
 	}
 
-	err = validateVaultHeaderValue(postHeadersMissing, requestURL, canaryHeaderValue)
+	err = validateVaultHeaderValue(http.MethodPost, postHeadersMissing, postRequestURL, canaryHeaderValue)
 	if err == nil {
 		t.Error("validated POST request with missing Vault header")
 	}
 
-	err = validateVaultHeaderValue(postHeadersInvalid, requestURL, canaryHeaderValue)
+	err = validateVaultHeaderValue(http.MethodPost, postHeadersInvalid, postRequestURL, canaryHeaderValue)
 	if err == nil {
 		t.Error("validated POST request with invalid Vault header value")
 	}
 
-	err = validateVaultHeaderValue(postHeadersUnsigned, requestURL, canaryHeaderValue)
+	err = validateVaultHeaderValue(http.MethodPost, postHeadersUnsigned, postRequestURL, canaryHeaderValue)
 	if err == nil {
 		t.Error("validated POST request with unsigned Vault header")
 	}
 
-	err = validateVaultHeaderValue(postHeadersValid, requestURL, canaryHeaderValue)
+	err = validateVaultHeaderValue(http.MethodPost, postHeadersValid, postRequestURL, canaryHeaderValue)
 	if err != nil {
 		t.Errorf("did NOT validate valid POST request: %v", err)
 	}
 
-	err = validateVaultHeaderValue(postHeadersSplit, requestURL, canaryHeaderValue)
+	err = validateLoginIamRequestUrl(http.MethodPost, postRequestURL)
+	if err != nil {
+		t.Errorf("did NOT validate valid POST request: %v", err)
+	}
+
+	err = validateVaultHeaderValue(http.MethodPost, postHeadersSplit, postRequestURL, canaryHeaderValue)
 	if err != nil {
 		t.Errorf("did NOT validate valid POST request with split Authorization header: %v", err)
 	}
 }
 
+// TestBackend_pathLogin_NoClientConfig covers logging in via IAM auth when the
+// client config does not exist. This is a regression test to cover potential
+// panics when referencing the potentially-nil config in the login handler. For
+// details see https://github.com/hashicorp/vault/issues/23361.
+func TestBackend_pathLogin_NoClientConfig(t *testing.T) {
+	storage := new(logical.InmemStorage)
+	config := logical.TestBackendConfig()
+	config.StorageView = storage
+	b, err := Backend(config)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	err = b.Setup(context.Background(), config)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Intentionally left out the client configuration
+
+	roleEntry := &awsRoleEntry{
+		RoleID:   "foo",
+		Version:  currentRoleStorageVersion,
+		AuthType: iamAuthType,
+	}
+	err = b.setRole(context.Background(), storage, testValidRoleName, roleEntry)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	loginData, err := defaultLoginData()
+	if err != nil {
+		t.Fatal(err)
+	}
+	loginRequest := &logical.Request{
+		Operation:  logical.UpdateOperation,
+		Path:       "login",
+		Storage:    storage,
+		Data:       loginData,
+		Connection: &logical.Connection{},
+	}
+	resp, err := b.HandleRequest(context.Background(), loginRequest)
+	if err != nil {
+		t.Fatalf("expected nil error, got: %v", err)
+	}
+	if !resp.IsError() {
+		t.Fatalf("expected error response, got: %+v", resp)
+	}
+}
+
 // TestBackend_pathLogin_IAMHeaders tests login with iam_request_headers,
 // supporting both base64 encoded string and JSON headers
 func TestBackend_pathLogin_IAMHeaders(t *testing.T) {
@@ -386,6 +565,106 @@ func TestBackend_pathLogin_IAMHeaders(t *testing.T) {
 	}
 }
 
+// TestBackend_pathLogin_IAMRoleResolution tests role resolution for an Iam login
+func TestBackend_pathLogin_IAMRoleResolution(t *testing.T) {
+	storage := &logical.InmemStorage{}
+	config := logical.TestBackendConfig()
+	config.StorageView = storage
+	b, err := Backend(config)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	err = b.Setup(context.Background(), config)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// sets up a test server to stand in for STS service
+	ts := setupIAMTestServer()
+	defer ts.Close()
+
+	clientConfigData := map[string]interface{}{
+		"iam_server_id_header_value": testVaultHeaderValue,
+		"sts_endpoint":               ts.URL,
+	}
+	clientRequest := &logical.Request{
+		Operation: logical.UpdateOperation,
+		Path:      "config/client",
+		Storage:   storage,
+		Data:      clientConfigData,
+	}
+	_, err = b.HandleRequest(context.Background(), clientRequest)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Configure identity.
+	_, err = b.HandleRequest(context.Background(), &logical.Request{
+		Operation: logical.UpdateOperation,
+		Path:      "config/identity",
+		Storage:   storage,
+		Data: map[string]interface{}{
+			"iam_alias": "role_id",
+			"iam_metadata": []string{
+				"account_id",
+				"auth_type",
+				"canonical_arn",
+				"client_arn",
+				"client_user_id",
+				"inferred_aws_region",
+				"inferred_entity_id",
+				"inferred_entity_type",
+			},
+			"ec2_alias": "role_id",
+			"ec2_metadata": []string{
+				"account_id",
+				"ami_id",
+				"instance_id",
+				"region",
+			},
+		},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// create a role entry
+	roleEntry := &awsRoleEntry{
+		RoleID:   "foo",
+		Version:  currentRoleStorageVersion,
+		AuthType: iamAuthType,
+	}
+
+	if err := b.setRole(context.Background(), storage, testValidRoleName, roleEntry); err != nil {
+		t.Fatalf("failed to set entry: %s", err)
+	}
+
+	// create a baseline loginData map structure, including iam_request_headers
+	// already base64encoded. This is the "Default" loginData used for all tests.
+	// Each sub test can override the map's iam_request_headers entry
+	loginData, err := defaultLoginData()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	loginRequest := &logical.Request{
+		Operation:  logical.ResolveRoleOperation,
+		Path:       "login",
+		Storage:    storage,
+		Data:       loginData,
+		Connection: &logical.Connection{},
+	}
+
+	resp, err := b.HandleRequest(context.Background(), loginRequest)
+	if err != nil || resp == nil || resp.IsError() {
+		t.Errorf("unexpected failed role resolution:\nresp: %#v\n\nerr: %v", resp, err)
+	}
+	if resp.Data["role"] != testValidRoleName {
+		t.Fatalf("Role was not as expected. Expected %s, received %s", testValidRoleName, resp.Data["role"])
+	}
+}
+
 func TestBackend_defaultAliasMetadata(t *testing.T) {
 	storage := &logical.InmemStorage{}
 	config := logical.TestBackendConfig()
@@ -522,6 +801,58 @@ func TestBackend_defaultAliasMetadata(t *testing.T) {
 	}
 }
 
+func TestRegionFromHeader(t *testing.T) {
+	tcs := map[string]struct {
+		header              string
+		expectedRegion      string
+		expectedSTSEndpoint string
+	}{
+		"us-east-1": {
+			header:              "AWS4-HMAC-SHA256 Credential=AAAAAAAAAAAAAAAAAAAA/20230719/us-east-1/sts/aws4_request, SignedHeaders=content-length;content-type;host;x-amz-date, Signature=aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+			expectedRegion:      "us-east-1",
+			expectedSTSEndpoint: "https://sts.us-east-1.amazonaws.com",
+		},
+		"us-west-2": {
+			header:              "AWS4-HMAC-SHA256 Credential=AAAAAAAAAAAAAAAAAAAA/20230719/us-west-2/sts/aws4_request, SignedHeaders=content-length;content-type;host;x-amz-date, Signature=aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+			expectedRegion:      "us-west-2",
+			expectedSTSEndpoint: "https://sts.us-west-2.amazonaws.com",
+		},
+		"ap-northeast-3": {
+			header:              "AWS4-HMAC-SHA256 Credential=AAAAAAAAAAAAAAAAAAAA/20230719/ap-northeast-3/sts/aws4_request, SignedHeaders=content-length;content-type;host;x-amz-date, Signature=aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+			expectedRegion:      "ap-northeast-3",
+			expectedSTSEndpoint: "https://sts.ap-northeast-3.amazonaws.com",
+		},
+		"us-gov-east-1": {
+			header:              "AWS4-HMAC-SHA256 Credential=AAAAAAAAAAAAAAAAAAAA/20230719/us-gov-east-1/sts/aws4_request, SignedHeaders=content-length;content-type;host;x-amz-date, Signature=aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+			expectedRegion:      "us-gov-east-1",
+			expectedSTSEndpoint: "https://sts.us-gov-east-1.amazonaws.com",
+		},
+	}
+	for name, tc := range tcs {
+		t.Run(name, func(t *testing.T) {
+			region, err := awsRegionFromHeader(tc.header)
+			assert.NoError(t, err)
+			assert.Equal(t, tc.expectedRegion, region)
+
+			stsEndpoint, err := stsRegionalEndpoint(region)
+			assert.NoError(t, err)
+			assert.Equal(t, tc.expectedSTSEndpoint, stsEndpoint)
+		})
+	}
+
+	t.Run("invalid-header", func(t *testing.T) {
+		region, err := awsRegionFromHeader("this-is-an-invalid-header/foobar")
+		assert.EqualError(t, err, "invalid header format")
+		assert.Empty(t, region)
+	})
+
+	t.Run("invalid-region", func(t *testing.T) {
+		endpoint, err := stsRegionalEndpoint("fake-region-1")
+		assert.EqualError(t, err, "unable to get regional STS endpoint for region: fake-region-1")
+		assert.Empty(t, endpoint)
+	})
+}
+
 func defaultLoginData() (map[string]interface{}, error) {
 	awsSession, err := session.NewSession()
 	if err != nil {
