@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package transit
 
 import (
@@ -8,6 +11,8 @@ import (
 	"strings"
 	"testing"
 
+	"github.com/hashicorp/vault/helper/constants"
+
 	"golang.org/x/crypto/ed25519"
 
 	"github.com/hashicorp/vault/sdk/helper/keysutil"
@@ -23,6 +28,7 @@ type signOutcome struct {
 	requestOk bool
 	valid     bool
 	keyValid  bool
+	reference string
 }
 
 func TestTransit_SignVerify_ECDSA(t *testing.T) {
@@ -55,7 +61,7 @@ func testTransit_SignVerify_ECDSA(t *testing.T, bits int) {
 	}
 
 	// Now, change the key value to something we control
-	p, _, err := b.lm.GetPolicy(context.Background(), keysutil.PolicyRequest{
+	p, _, err := b.GetPolicy(context.Background(), keysutil.PolicyRequest{
 		Storage: storage,
 		Name:    "foo",
 	}, b.GetRandomReader())
@@ -152,7 +158,7 @@ func testTransit_SignVerify_ECDSA(t *testing.T, bits int) {
 		req.Path = "sign/foo" + postpath
 		resp, err := b.HandleRequest(context.Background(), req)
 		if err != nil && !errExpected {
-			t.Fatal(err)
+			t.Fatalf("request: %v\nerror: %v", req, err)
 		}
 		if resp == nil {
 			t.Fatal("expected non-nil response")
@@ -237,6 +243,26 @@ func testTransit_SignVerify_ECDSA(t *testing.T, bits int) {
 	sig = signRequest(req, false, "")
 	verifyRequest(req, false, "", sig)
 
+	req.Data["hash_algorithm"] = "sha2-512"
+	sig = signRequest(req, false, "")
+	verifyRequest(req, false, "", sig)
+
+	req.Data["hash_algorithm"] = "sha3-224"
+	sig = signRequest(req, false, "")
+	verifyRequest(req, false, "", sig)
+
+	req.Data["hash_algorithm"] = "sha3-256"
+	sig = signRequest(req, false, "")
+	verifyRequest(req, false, "", sig)
+
+	req.Data["hash_algorithm"] = "sha3-384"
+	sig = signRequest(req, false, "")
+	verifyRequest(req, false, "", sig)
+
+	req.Data["hash_algorithm"] = "sha3-512"
+	sig = signRequest(req, false, "")
+	verifyRequest(req, false, "", sig)
+
 	req.Data["prehashed"] = true
 	sig = signRequest(req, false, "")
 	verifyRequest(req, false, "", sig)
@@ -377,7 +403,7 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 	}
 
 	// Get the keys for later
-	fooP, _, err := b.lm.GetPolicy(context.Background(), keysutil.PolicyRequest{
+	fooP, _, err := b.GetPolicy(context.Background(), keysutil.PolicyRequest{
 		Storage: storage,
 		Name:    "foo",
 	}, b.GetRandomReader())
@@ -385,7 +411,7 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	barP, _, err := b.lm.GetPolicy(context.Background(), keysutil.PolicyRequest{
+	barP, _, err := b.GetPolicy(context.Background(), keysutil.PolicyRequest{
 		Storage: storage,
 		Name:    "bar",
 	}, b.GetRandomReader())
@@ -461,6 +487,7 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 			}
 			for i, v := range sig {
 				batchRequestItems[i]["signature"] = v
+				batchRequestItems[i]["reference"] = outcome[i].reference
 			}
 		} else if attachSig {
 			req.Data["signature"] = sig[0]
@@ -513,6 +540,9 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 				if pubKeyRaw, ok := req.Data["public_key"]; ok {
 					validatePublicKey(t, batchRequestItems[i]["input"], sig[i], pubKeyRaw.([]byte), outcome[i].keyValid, postpath, b)
 				}
+				if v.Reference != outcome[i].reference {
+					t.Fatalf("verification failed, mismatched references %s vs %s", v.Reference, outcome[i].reference)
+				}
 			}
 			return
 		}
@@ -612,15 +642,18 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 
 	// Test Batch Signing
 	batchInput := []batchRequestSignItem{
-		{"context": "abcd", "input": "dGhlIHF1aWNrIGJyb3duIGZveA=="},
-		{"context": "efgh", "input": "dGhlIHF1aWNrIGJyb3duIGZveA=="},
+		{"context": "abcd", "input": "dGhlIHF1aWNrIGJyb3duIGZveA==", "reference": "uno"},
+		{"context": "efgh", "input": "dGhlIHF1aWNrIGJyb3duIGZveA==", "reference": "dos"},
 	}
 
 	req.Data = map[string]interface{}{
 		"batch_input": batchInput,
 	}
 
-	outcome = []signOutcome{{requestOk: true, valid: true, keyValid: true}, {requestOk: true, valid: true, keyValid: true}}
+	outcome = []signOutcome{
+		{requestOk: true, valid: true, keyValid: true, reference: "uno"},
+		{requestOk: true, valid: true, keyValid: true, reference: "dos"},
+	}
 
 	sig = signRequest(req, false, "foo")
 	verifyRequest(req, false, outcome, "foo", sig, true)
@@ -680,3 +713,269 @@ func TestTransit_SignVerify_ED25519(t *testing.T) {
 	outcome[1].valid = false
 	verifyRequest(req, false, outcome, "bar", goodsig, true)
 }
+
+func TestTransit_SignVerify_RSA_PSS(t *testing.T) {
+	t.Run("2048", func(t *testing.T) {
+		testTransit_SignVerify_RSA_PSS(t, 2048)
+	})
+	t.Run("3072", func(t *testing.T) {
+		testTransit_SignVerify_RSA_PSS(t, 3072)
+	})
+	t.Run("4096", func(t *testing.T) {
+		testTransit_SignVerify_RSA_PSS(t, 4096)
+	})
+}
+
+func testTransit_SignVerify_RSA_PSS(t *testing.T, bits int) {
+	b, storage := createBackendWithSysView(t)
+
+	// First create a key
+	req := &logical.Request{
+		Storage:   storage,
+		Operation: logical.UpdateOperation,
+		Path:      "keys/foo",
+		Data: map[string]interface{}{
+			"type": fmt.Sprintf("rsa-%d", bits),
+		},
+	}
+	_, err := b.HandleRequest(context.Background(), req)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	signRequest := func(errExpected bool, postpath string) string {
+		t.Helper()
+		req.Path = "sign/foo" + postpath
+		resp, err := b.HandleRequest(context.Background(), req)
+		if err != nil && !errExpected {
+			t.Fatal(err)
+		}
+		if resp == nil {
+			t.Fatal("expected non-nil response")
+		}
+		if errExpected {
+			if !resp.IsError() {
+				t.Fatalf("bad: should have gotten error response: %#v", *resp)
+			}
+			return ""
+		}
+		if resp.IsError() {
+			t.Fatalf("bad: got error response: %#v", *resp)
+		}
+		// Since we are reusing the same request, let's clear the salt length each time.
+		delete(req.Data, "salt_length")
+
+		value, ok := resp.Data["signature"]
+		if !ok {
+			t.Fatalf("no signature key found in returned data, got resp data %#v", resp.Data)
+		}
+		return value.(string)
+	}
+
+	verifyRequest := func(errExpected bool, postpath, sig string) {
+		t.Helper()
+		req.Path = "verify/foo" + postpath
+		req.Data["signature"] = sig
+		resp, err := b.HandleRequest(context.Background(), req)
+		if err != nil {
+			if errExpected {
+				return
+			}
+			t.Fatalf("got error: %v, sig was %v", err, sig)
+		}
+		if resp == nil {
+			t.Fatal("expected non-nil response")
+		}
+		if resp.IsError() {
+			if errExpected {
+				return
+			}
+			t.Fatalf("bad: got error response: %#v", *resp)
+		}
+		value, ok := resp.Data["valid"]
+		if !ok {
+			t.Fatalf("no valid key found in returned data, got resp data %#v", resp.Data)
+		}
+		if !value.(bool) && !errExpected {
+			t.Fatalf("verification failed; req was %#v, resp is %#v", *req, *resp)
+		} else if value.(bool) && errExpected {
+			t.Fatalf("expected error and didn't get one; req was %#v, resp is %#v", *req, *resp)
+		}
+		// Since we are reusing the same request, let's clear the signature each time.
+		delete(req.Data, "signature")
+	}
+
+	newReqData := func(hashAlgorithm string, marshalingName string) map[string]interface{} {
+		return map[string]interface{}{
+			"input":                "dGhlIHF1aWNrIGJyb3duIGZveA==",
+			"signature_algorithm":  "pss",
+			"hash_algorithm":       hashAlgorithm,
+			"marshaling_algorithm": marshalingName,
+		}
+	}
+
+	signAndVerifyRequest := func(hashAlgorithm string, marshalingName string, signSaltLength string, signErrExpected bool, verifySaltLength string, verifyErrExpected bool) {
+		t.Log("\t\t\t", signSaltLength, "/", verifySaltLength)
+		req.Data = newReqData(hashAlgorithm, marshalingName)
+
+		req.Data["salt_length"] = signSaltLength
+		t.Log("\t\t\t\t", "sign req data:", req.Data)
+		sig := signRequest(signErrExpected, "")
+
+		req.Data["salt_length"] = verifySaltLength
+		t.Log("\t\t\t\t", "verify req data:", req.Data)
+		verifyRequest(verifyErrExpected, "", sig)
+	}
+
+	invalidSaltLengths := []string{"bar", "-2"}
+	t.Log("invalidSaltLengths:", invalidSaltLengths)
+
+	autoSaltLengths := []string{"auto", "0"}
+	t.Log("autoSaltLengths:", autoSaltLengths)
+
+	hashSaltLengths := []string{"hash", "-1"}
+	t.Log("hashSaltLengths:", hashSaltLengths)
+
+	positiveSaltLengths := []string{"1"}
+	t.Log("positiveSaltLengths:", positiveSaltLengths)
+
+	nonAutoSaltLengths := append(hashSaltLengths, positiveSaltLengths...)
+	t.Log("nonAutoSaltLengths:", nonAutoSaltLengths)
+
+	validSaltLengths := append(autoSaltLengths, nonAutoSaltLengths...)
+	t.Log("validSaltLengths:", validSaltLengths)
+
+	testCombinatorics := func(t *testing.T, hashAlgorithm string, marshalingName string) {
+		t.Log("\t\t", "valid", "/", "invalid salt lengths")
+		for _, validSaltLength := range validSaltLengths {
+			for _, invalidSaltLength := range invalidSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, validSaltLength, false, invalidSaltLength, true)
+			}
+		}
+
+		t.Log("\t\t", "invalid", "/", "invalid salt lengths")
+		for _, invalidSaltLength1 := range invalidSaltLengths {
+			for _, invalidSaltLength2 := range invalidSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, invalidSaltLength1, true, invalidSaltLength2, true)
+			}
+		}
+
+		t.Log("\t\t", "invalid", "/", "valid salt lengths")
+		for _, invalidSaltLength := range invalidSaltLengths {
+			for _, validSaltLength := range validSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, invalidSaltLength, true, validSaltLength, true)
+			}
+		}
+
+		t.Log("\t\t", "valid", "/", "valid salt lengths")
+		for _, validSaltLength := range validSaltLengths {
+			signAndVerifyRequest(hashAlgorithm, marshalingName, validSaltLength, false, validSaltLength, false)
+		}
+
+		t.Log("\t\t", "hash", "/", "hash salt lengths")
+		for _, hashSaltLength1 := range hashSaltLengths {
+			for _, hashSaltLength2 := range hashSaltLengths {
+				if hashSaltLength1 != hashSaltLength2 {
+					signAndVerifyRequest(hashAlgorithm, marshalingName, hashSaltLength1, false, hashSaltLength2, false)
+				}
+			}
+		}
+
+		t.Log("\t\t", "hash", "/", "positive salt lengths")
+		for _, hashSaltLength := range hashSaltLengths {
+			for _, positiveSaltLength := range positiveSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, hashSaltLength, false, positiveSaltLength, true)
+			}
+		}
+
+		t.Log("\t\t", "positive", "/", "hash salt lengths")
+		for _, positiveSaltLength := range positiveSaltLengths {
+			for _, hashSaltLength := range hashSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, positiveSaltLength, false, hashSaltLength, true)
+			}
+		}
+
+		t.Log("\t\t", "auto", "/", "auto salt lengths")
+		for _, autoSaltLength1 := range autoSaltLengths {
+			for _, autoSaltLength2 := range autoSaltLengths {
+				if autoSaltLength1 != autoSaltLength2 {
+					signAndVerifyRequest(hashAlgorithm, marshalingName, autoSaltLength1, false, autoSaltLength2, false)
+				}
+			}
+		}
+
+		t.Log("\t\t", "auto", "/", "non-auto salt lengths")
+		for _, autoSaltLength := range autoSaltLengths {
+			for _, nonAutoSaltLength := range nonAutoSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, autoSaltLength, false, nonAutoSaltLength, true)
+			}
+		}
+
+		t.Log("\t\t", "non-auto", "/", "auto salt lengths")
+		for _, nonAutoSaltLength := range nonAutoSaltLengths {
+			for _, autoSaltLength := range autoSaltLengths {
+				signAndVerifyRequest(hashAlgorithm, marshalingName, nonAutoSaltLength, false, autoSaltLength, false)
+			}
+		}
+	}
+
+	testAutoSignAndVerify := func(t *testing.T, hashAlgorithm string, marshalingName string) {
+		t.Log("\t\t", "Make a signature with an implicit, automatic salt length")
+		req.Data = newReqData(hashAlgorithm, marshalingName)
+		t.Log("\t\t\t", "sign req data:", req.Data)
+		sig := signRequest(false, "")
+
+		t.Log("\t\t", "Verify it with an implicit, automatic salt length")
+		t.Log("\t\t\t", "verify req data:", req.Data)
+		verifyRequest(false, "", sig)
+
+		t.Log("\t\t", "Verify it with an explicit, automatic salt length")
+		for _, autoSaltLength := range autoSaltLengths {
+			t.Log("\t\t\t", "auto", "/", autoSaltLength)
+			req.Data["salt_length"] = autoSaltLength
+			t.Log("\t\t\t\t", "verify req data:", req.Data)
+			verifyRequest(false, "", sig)
+		}
+
+		t.Log("\t\t", "Try to verify it with an explicit, incorrect salt length")
+		for _, nonAutoSaltLength := range nonAutoSaltLengths {
+			t.Log("\t\t\t", "auto", "/", nonAutoSaltLength)
+			req.Data["salt_length"] = nonAutoSaltLength
+			t.Log("\t\t\t\t", "verify req data:", req.Data)
+			verifyRequest(true, "", sig)
+		}
+
+		t.Log("\t\t", "Make a signature with an explicit, valid salt length & and verify it with an implicit, automatic salt length")
+		for _, validSaltLength := range validSaltLengths {
+			t.Log("\t\t\t", validSaltLength, "/", "auto")
+
+			req.Data = newReqData(hashAlgorithm, marshalingName)
+			req.Data["salt_length"] = validSaltLength
+			t.Log("\t\t\t", "sign req data:", req.Data)
+			sig := signRequest(false, "")
+
+			t.Log("\t\t\t", "verify req data:", req.Data)
+			verifyRequest(false, "", sig)
+		}
+	}
+
+	for hashAlgorithm := range keysutil.HashTypeMap {
+		t.Log("Hash algorithm:", hashAlgorithm)
+		if hashAlgorithm == "none" {
+			continue
+		}
+
+		for marshalingName := range keysutil.MarshalingTypeMap {
+			t.Log("\t", "Marshaling type:", marshalingName)
+			testName := fmt.Sprintf("%s-%s", hashAlgorithm, marshalingName)
+			t.Run(testName, func(t *testing.T) {
+				if constants.IsFIPS() && strings.HasPrefix(hashAlgorithm, "sha3-") {
+					t.Skip("\t", "Skipping hashing algo on fips:", hashAlgorithm)
+				}
+
+				testCombinatorics(t, hashAlgorithm, marshalingName)
+				testAutoSignAndVerify(t, hashAlgorithm, marshalingName)
+			})
+		}
+	}
+}
