@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package http
 
 import (
@@ -6,14 +9,14 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
+	"mime"
 	"net"
 	"net/http"
 	"strconv"
 	"strings"
 	"time"
 
-	"github.com/hashicorp/errwrap"
-	uuid "github.com/hashicorp/go-uuid"
+	"github.com/hashicorp/go-uuid"
 	"github.com/hashicorp/vault/helper/namespace"
 	"github.com/hashicorp/vault/sdk/helper/consts"
 	"github.com/hashicorp/vault/sdk/logical"
@@ -39,6 +42,8 @@ func (b *bufferedReader) Close() error {
 	return b.rOrig.Close()
 }
 
+const MergePatchContentTypeHeader = "application/merge-patch+json"
+
 func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.Request) (*logical.Request, io.ReadCloser, int, error) {
 	ns, err := namespace.FromContext(r.Context())
 	if err != nil {
@@ -69,6 +74,7 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 				return nil, nil, http.StatusBadRequest, nil
 			}
 			if list {
+				queryVals.Del("list")
 				op = logical.ListOperation
 				if !strings.HasSuffix(path, "/") {
 					path += "/"
@@ -76,9 +82,7 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 			}
 		}
 
-		if !list {
-			data = parseQuery(queryVals)
-		}
+		data = parseQuery(queryVals)
 
 		switch {
 		case strings.HasPrefix(path, "sys/pprof/"):
@@ -86,6 +90,8 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 			responseWriter = w
 		case path == "sys/storage/raft/snapshot":
 			responseWriter = w
+		case path == "sys/internal/counters/activity/export":
+			responseWriter = w
 		case path == "sys/monitor":
 			passHTTPReq = true
 			responseWriter = w
@@ -99,10 +105,11 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 		bufferedBody := newBufferedReader(r.Body)
 		r.Body = bufferedBody
 
-		// If we are uploading a snapshot we don't want to parse it. Instead
-		// we will simply add the HTTP request to the logical request object
-		// for later consumption.
-		if path == "sys/storage/raft/snapshot" || path == "sys/storage/raft/snapshot-force" {
+		// If we are uploading a snapshot or receiving an ocsp-request (which
+		// is der encoded) we don't want to parse it. Instead, we will simply
+		// add the HTTP request to the logical request object for later consumption.
+		contentType := r.Header.Get("Content-Type")
+		if path == "sys/storage/raft/snapshot" || path == "sys/storage/raft/snapshot-force" || isOcspRequest(contentType) {
 			passHTTPReq = true
 			origBody = r.Body
 		} else {
@@ -112,13 +119,17 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 			// and an incorrect content-type).
 			head, err := bufferedBody.Peek(512)
 			if err != nil && err != bufio.ErrBufferFull && err != io.EOF {
-				return nil, nil, http.StatusBadRequest, err
+				status := http.StatusBadRequest
+				logical.AdjustErrorStatusCode(&status, err)
+				return nil, nil, status, fmt.Errorf("error reading data")
 			}
 
-			if isForm(head, r.Header.Get("Content-Type")) {
+			if isForm(head, contentType) {
 				formData, err := parseFormRequest(r)
 				if err != nil {
-					return nil, nil, http.StatusBadRequest, fmt.Errorf("error parsing form data: %w", err)
+					status := http.StatusBadRequest
+					logical.AdjustErrorStatusCode(&status, err)
+					return nil, nil, status, fmt.Errorf("error parsing form data")
 				}
 
 				data = formData
@@ -129,25 +140,59 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 					err = nil
 				}
 				if err != nil {
-					return nil, nil, http.StatusBadRequest, err
+					status := http.StatusBadRequest
+					logical.AdjustErrorStatusCode(&status, err)
+					return nil, nil, status, fmt.Errorf("error parsing JSON")
 				}
 			}
 		}
 
+	case "PATCH":
+		op = logical.PatchOperation
+
+		contentTypeHeader := r.Header.Get("Content-Type")
+		contentType, _, err := mime.ParseMediaType(contentTypeHeader)
+		if err != nil {
+			status := http.StatusBadRequest
+			logical.AdjustErrorStatusCode(&status, err)
+			return nil, nil, status, err
+		}
+
+		if contentType != MergePatchContentTypeHeader {
+			return nil, nil, http.StatusUnsupportedMediaType, fmt.Errorf("PATCH requires Content-Type of %s, provided %s", MergePatchContentTypeHeader, contentType)
+		}
+
+		origBody, err = parseJSONRequest(perfStandby, r, w, &data)
+
+		if err == io.EOF {
+			data = nil
+			err = nil
+		}
+
+		if err != nil {
+			status := http.StatusBadRequest
+			logical.AdjustErrorStatusCode(&status, err)
+			return nil, nil, status, fmt.Errorf("error parsing JSON")
+		}
+
 	case "LIST":
 		op = logical.ListOperation
 		if !strings.HasSuffix(path, "/") {
 			path += "/"
 		}
 
-	case "OPTIONS", "HEAD":
+		data = parseQuery(r.URL.Query())
+	case "HEAD":
+		op = logical.HeaderOperation
+		data = parseQuery(r.URL.Query())
+	case "OPTIONS":
 	default:
 		return nil, nil, http.StatusMethodNotAllowed, nil
 	}
 
 	requestId, err := uuid.GenerateUUID()
 	if err != nil {
-		return nil, nil, http.StatusBadRequest, errwrap.Wrapf("failed to generate identifier for the request: {{err}}", err)
+		return nil, nil, http.StatusInternalServerError, fmt.Errorf("failed to generate identifier for the request: %w", err)
 	}
 
 	req := &logical.Request{
@@ -169,6 +214,15 @@ func buildLogicalRequestNoAuth(perfStandby bool, w http.ResponseWriter, r *http.
 	return req, origBody, 0, nil
 }
 
+func isOcspRequest(contentType string) bool {
+	contentType, _, err := mime.ParseMediaType(contentType)
+	if err != nil {
+		return false
+	}
+
+	return contentType == "application/ocsp-request"
+}
+
 func buildLogicalPath(r *http.Request) (string, int, error) {
 	ns, err := namespace.FromContext(r.Context())
 	if err != nil {
@@ -208,32 +262,29 @@ func buildLogicalPath(r *http.Request) (string, int, error) {
 	return path, 0, nil
 }
 
-func buildLogicalRequest(core *vault.Core, w http.ResponseWriter, r *http.Request) (*logical.Request, io.ReadCloser, int, error) {
+func buildLogicalRequest(core *vault.Core, w http.ResponseWriter, r *http.Request, chrootNamespace string) (*logical.Request, io.ReadCloser, int, error) {
 	req, origBody, status, err := buildLogicalRequestNoAuth(core.PerfStandby(), w, r)
 	if err != nil || status != 0 {
 		return nil, nil, status, err
 	}
-	req, err = requestAuth(core, r, req)
-	if err != nil {
-		if errwrap.Contains(err, logical.ErrPermissionDenied.Error()) {
-			return nil, nil, http.StatusForbidden, nil
-		}
-		return nil, nil, http.StatusBadRequest, errwrap.Wrapf("error performing token check: {{err}}", err)
-	}
+	req.ChrootNamespace = chrootNamespace
+
+	req.SetRequiredState(r.Header.Values(VaultIndexHeaderName))
+	requestAuth(r, req)
 
 	req, err = requestWrapInfo(r, req)
 	if err != nil {
-		return nil, nil, http.StatusBadRequest, errwrap.Wrapf("error parsing X-Vault-Wrap-TTL header: {{err}}", err)
+		return nil, nil, http.StatusBadRequest, fmt.Errorf("error parsing X-Vault-Wrap-TTL header: %w", err)
 	}
 
 	err = parseMFAHeader(req)
 	if err != nil {
-		return nil, nil, http.StatusBadRequest, errwrap.Wrapf("failed to parse X-Vault-MFA header: {{err}}", err)
+		return nil, nil, http.StatusBadRequest, fmt.Errorf("failed to parse X-Vault-MFA header: %w", err)
 	}
 
 	err = requestPolicyOverride(r, req)
 	if err != nil {
-		return nil, nil, http.StatusBadRequest, errwrap.Wrapf(fmt.Sprintf(`failed to parse %s header: {{err}}`, PolicyOverrideHeaderName), err)
+		return nil, nil, http.StatusBadRequest, fmt.Errorf("failed to parse %s header: %w", PolicyOverrideHeaderName, err)
 	}
 
 	return req, origBody, 0, nil
@@ -242,25 +293,25 @@ func buildLogicalRequest(core *vault.Core, w http.ResponseWriter, r *http.Reques
 // handleLogical returns a handler for processing logical requests. These requests
 // may or may not end up getting forwarded under certain scenarios if the node
 // is a performance standby. Some of these cases include:
-//     - Perf standby and token with limited use count.
-//     - Perf standby and token re-validation needed (e.g. due to invalid token).
-//     - Perf standby and control group error.
-func handleLogical(core *vault.Core) http.Handler {
-	return handleLogicalInternal(core, false, false)
+//   - Perf standby and token with limited use count.
+//   - Perf standby and token re-validation needed (e.g. due to invalid token).
+//   - Perf standby and control group error.
+func handleLogical(core *vault.Core, chrootNamespace string) http.Handler {
+	return handleLogicalInternal(core, false, false, chrootNamespace)
 }
 
 // handleLogicalWithInjector returns a handler for processing logical requests
 // that also have their logical response data injected at the top-level payload.
 // All forwarding behavior remains the same as `handleLogical`.
-func handleLogicalWithInjector(core *vault.Core) http.Handler {
-	return handleLogicalInternal(core, true, false)
+func handleLogicalWithInjector(core *vault.Core, chrootNamespace string) http.Handler {
+	return handleLogicalInternal(core, true, false, chrootNamespace)
 }
 
 // handleLogicalNoForward returns a handler for processing logical local-only
 // requests. These types of requests never forwarded, and return an
 // `vault.ErrCannotForwardLocalOnly` error if attempted to do so.
-func handleLogicalNoForward(core *vault.Core) http.Handler {
-	return handleLogicalInternal(core, false, true)
+func handleLogicalNoForward(core *vault.Core, chrootNamespace string) http.Handler {
+	return handleLogicalInternal(core, false, true, chrootNamespace)
 }
 
 func handleLogicalRecovery(raw *vault.RawBackend, token *atomic.String) http.Handler {
@@ -293,135 +344,33 @@ func handleLogicalRecovery(raw *vault.RawBackend, token *atomic.String) http.Han
 // handleLogicalInternal is a common helper that returns a handler for
 // processing logical requests. The behavior depends on the various boolean
 // toggles. Refer to usage on functions for possible behaviors.
-func handleLogicalInternal(core *vault.Core, injectDataIntoTopLevel bool, noForward bool) http.Handler {
+func handleLogicalInternal(core *vault.Core, injectDataIntoTopLevel bool, noForward bool, chrootNamespace string) http.Handler {
 	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
-		req, origBody, statusCode, err := buildLogicalRequest(core, w, r)
+		req, origBody, statusCode, err := buildLogicalRequest(core, w, r, chrootNamespace)
 		if err != nil || statusCode != 0 {
 			respondError(w, statusCode, err)
 			return
 		}
 
-		// Always forward requests that are using a limited use count token.
-		if core.PerfStandby() && req.ClientTokenRemainingUses > 0 {
-			// Prevent forwarding on local-only requests.
-			if noForward {
-				respondError(w, http.StatusBadRequest, vault.ErrCannotForwardLocalOnly)
-				return
-			}
-			if origBody != nil {
-				r.Body = origBody
-			}
-			forwardRequest(core, w, r)
+		// Websockets need to be handled at HTTP layer instead of logical requests.
+		ns, err := namespace.FromContext(r.Context())
+		if err != nil {
+			respondError(w, http.StatusInternalServerError, err)
 			return
 		}
-
-		// req.Path will be relative by this point. The prefix check is first
-		// to fail faster if we're not in this situation since it's a hot path
-		switch {
-		case strings.HasPrefix(req.Path, "sys/wrapping/"), strings.HasPrefix(req.Path, "auth/token/"):
-			// Get the token ns info; if we match the paths below we want to
-			// swap in the token context (but keep the relative path)
-			te := req.TokenEntry()
-			newCtx := r.Context()
-			if te != nil {
-				ns, err := vault.NamespaceByID(newCtx, te.NamespaceID, core)
-				if err != nil {
-					core.Logger().Warn("error looking up namespace from the token's namespace ID", "error", err)
-					respondError(w, http.StatusInternalServerError, err)
-					return
-				}
-				if ns != nil {
-					newCtx = namespace.ContextWithNamespace(newCtx, ns)
-				}
-			}
-			switch req.Path {
-			// Route the token wrapping request to its respective sys NS
-			case "sys/wrapping/lookup", "sys/wrapping/rewrap", "sys/wrapping/unwrap":
-				r = r.WithContext(newCtx)
-				if err := wrappingVerificationFunc(r.Context(), core, req); err != nil {
-					if errwrap.Contains(err, logical.ErrPermissionDenied.Error()) {
-						respondError(w, http.StatusForbidden, err)
-					} else {
-						respondError(w, http.StatusBadRequest, err)
-					}
-					return
-				}
-
-			// The -self paths have no meaning outside of the token NS, so
-			// requests for these paths always go to the token NS
-			case "auth/token/lookup-self", "auth/token/renew-self", "auth/token/revoke-self":
-				r = r.WithContext(newCtx)
-
-			// For the following operations, we can set the proper namespace context
-			// using the token's embedded nsID if a relative path was provided. Since
-			// this is done at the HTTP layer, the operation will still be gated by
-			// ACLs.
-			case "auth/token/lookup", "auth/token/renew", "auth/token/revoke", "auth/token/revoke-orphan":
-				token, ok := req.Data["token"]
-				// If the token is not present (e.g. a bad request), break out and let the backend
-				// handle the error
-				if !ok {
-					// If this is a token lookup request and if the token is not
-					// explicitly provided, it will use the client token so we simply set
-					// the context to the client token's context.
-					if req.Path == "auth/token/lookup" {
-						r = r.WithContext(newCtx)
-					}
-					break
-				}
-				_, nsID := namespace.SplitIDFromString(token.(string))
-				if nsID != "" {
-					ns, err := vault.NamespaceByID(newCtx, nsID, core)
-					if err != nil {
-						core.Logger().Warn("error looking up namespace from the token's namespace ID", "error", err)
-						respondError(w, http.StatusInternalServerError, err)
-						return
-					}
-					if ns != nil {
-						newCtx = namespace.ContextWithNamespace(newCtx, ns)
-						r = r.WithContext(newCtx)
-					}
-				}
-			}
-
-		// The following relative sys/leases/ paths handles re-routing requests
-		// to the proper namespace using the lease ID on applicable paths.
-		case strings.HasPrefix(req.Path, "sys/leases/"):
-			switch req.Path {
-			// For the following operations, we can set the proper namespace context
-			// using the lease's embedded nsID if a relative path was provided. Since
-			// this is done at the HTTP layer, the operation will still be gated by
-			// ACLs.
-			case "sys/leases/lookup", "sys/leases/renew", "sys/leases/revoke", "sys/leases/revoke-force":
-				leaseID, ok := req.Data["lease_id"]
-				// If lease ID is not present, break out and let the backend handle the error
-				if !ok {
-					break
-				}
-				_, nsID := namespace.SplitIDFromString(leaseID.(string))
-				if nsID != "" {
-					newCtx := r.Context()
-					ns, err := vault.NamespaceByID(newCtx, nsID, core)
-					if err != nil {
-						core.Logger().Warn("error looking up namespace from the lease's namespace ID", "error", err)
-						respondError(w, http.StatusInternalServerError, err)
-						return
-					}
-					if ns != nil {
-						newCtx = namespace.ContextWithNamespace(newCtx, ns)
-						r = r.WithContext(newCtx)
-					}
-				}
-			}
-
-		// Prevent any metrics requests to be forwarded from a standby node.
-		// Instead, we return an error since we cannot be sure if we have an
-		// active token store to validate the provided token.
-		case strings.HasPrefix(req.Path, "sys/metrics"):
-			if isStandby, _ := core.Standby(); isStandby {
-				respondError(w, http.StatusBadRequest, vault.ErrCannotForwardLocalOnly)
-				return
-			}
+		nsPath := ns.Path
+		if ns.ID == namespace.RootNamespaceID {
+			nsPath = ""
+		}
+		if strings.HasPrefix(r.URL.Path, fmt.Sprintf("/v1/%ssys/events/subscribe/", nsPath)) {
+			handler := handleEventsSubscribe(core, req)
+			handler.ServeHTTP(w, r)
+			return
+		}
+		handler := handleEntPaths(nsPath, core, r)
+		if handler != nil {
+			handler.ServeHTTP(w, r)
+			return
 		}
 
 		// Make the internal request. We attach the connection info
@@ -447,13 +396,13 @@ func handleLogicalInternal(core *vault.Core, injectDataIntoTopLevel bool, noForw
 			return
 		default:
 			// Build and return the proper response if everything is fine.
-			respondLogical(w, r, req, resp, injectDataIntoTopLevel)
+			respondLogical(core, w, r, req, resp, injectDataIntoTopLevel)
 			return
 		}
 	})
 }
 
-func respondLogical(w http.ResponseWriter, r *http.Request, req *logical.Request, resp *logical.Response, injectDataIntoTopLevel bool) {
+func respondLogical(core *vault.Core, w http.ResponseWriter, r *http.Request, req *logical.Request, resp *logical.Response, injectDataIntoTopLevel bool) {
 	var httpResp *logical.HTTPResponse
 	var ret interface{}
 
@@ -503,6 +452,8 @@ func respondLogical(w http.ResponseWriter, r *http.Request, req *logical.Request
 		}
 	}
 
+	adjustResponse(core, w, req)
+
 	// Respond
 	respondOk(w, ret)
 	return
@@ -599,10 +550,18 @@ WRITE_RESPONSE:
 		w.Header().Set("Content-Type", contentType)
 	}
 
-	if cacheControl, ok := resp.Data[logical.HTTPRawCacheControl].(string); ok {
+	if cacheControl, ok := resp.Data[logical.HTTPCacheControlHeader].(string); ok {
 		w.Header().Set("Cache-Control", cacheControl)
 	}
 
+	if pragma, ok := resp.Data[logical.HTTPPragmaHeader].(string); ok {
+		w.Header().Set("Pragma", pragma)
+	}
+
+	if wwwAuthn, ok := resp.Data[logical.HTTPWWWAuthenticateHeader].(string); ok {
+		w.Header().Set("WWW-Authenticate", wwwAuthn)
+	}
+
 	w.WriteHeader(status)
 	w.Write(body)
 }
@@ -611,14 +570,21 @@ WRITE_RESPONSE:
 // attaching to a logical request
 func getConnection(r *http.Request) (connection *logical.Connection) {
 	var remoteAddr string
+	var remotePort int
 
-	remoteAddr, _, err := net.SplitHostPort(r.RemoteAddr)
+	remoteAddr, port, err := net.SplitHostPort(r.RemoteAddr)
 	if err != nil {
 		remoteAddr = ""
+	} else {
+		remotePort, err = strconv.Atoi(port)
+		if err != nil {
+			remotePort = 0
+		}
 	}
 
 	connection = &logical.Connection{
 		RemoteAddr: remoteAddr,
+		RemotePort: remotePort,
 		ConnState:  r.TLS,
 	}
 	return
