@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package raft
 
 import (
@@ -15,7 +18,7 @@ import (
 	"testing"
 	"time"
 
-	hclog "github.com/hashicorp/go-hclog"
+	"github.com/hashicorp/go-hclog"
 	"github.com/hashicorp/raft"
 	"github.com/hashicorp/vault/sdk/logical"
 	"github.com/hashicorp/vault/sdk/physical"
@@ -55,7 +58,7 @@ func addPeer(t *testing.T, leader, follower *RaftBackend) {
 }
 
 func TestRaft_Snapshot_Loading(t *testing.T) {
-	raft, dir := getRaft(t, true, false)
+	raft, dir := GetRaft(t, true, false)
 	defer os.RemoveAll(dir)
 
 	// Write some data
@@ -136,11 +139,10 @@ func TestRaft_Snapshot_Loading(t *testing.T) {
 	if !bytes.Equal(computed1, computed3) {
 		t.Fatal("hashes did not match")
 	}
-
 }
 
 func TestRaft_Snapshot_Index(t *testing.T) {
-	raft, dir := getRaft(t, true, false)
+	raft, dir := GetRaft(t, true, false)
 	defer os.RemoveAll(dir)
 
 	err := raft.Put(context.Background(), &physical.Entry{
@@ -227,9 +229,9 @@ func TestRaft_Snapshot_Index(t *testing.T) {
 }
 
 func TestRaft_Snapshot_Peers(t *testing.T) {
-	raft1, dir := getRaft(t, true, false)
-	raft2, dir2 := getRaft(t, false, false)
-	raft3, dir3 := getRaft(t, false, false)
+	raft1, dir := GetRaft(t, true, false)
+	raft2, dir2 := GetRaft(t, false, false)
+	raft3, dir3 := GetRaft(t, false, false)
 	defer os.RemoveAll(dir)
 	defer os.RemoveAll(dir2)
 	defer os.RemoveAll(dir3)
@@ -310,9 +312,9 @@ func ensureCommitApplied(t *testing.T, leaderCommitIdx uint64, backend *RaftBack
 }
 
 func TestRaft_Snapshot_Restart(t *testing.T) {
-	raft1, dir := getRaft(t, true, false)
+	raft1, dir := GetRaft(t, true, false)
 	defer os.RemoveAll(dir)
-	raft2, dir2 := getRaft(t, false, false)
+	raft2, dir2 := GetRaft(t, false, false)
 	defer os.RemoveAll(dir2)
 
 	// Write some data
@@ -350,7 +352,7 @@ func TestRaft_Snapshot_Restart(t *testing.T) {
 		t.Fatal(peers)
 	}
 
-	// Shutdown raft1
+	// Finalize raft1
 	if err := raft1.TeardownCluster(nil); err != nil {
 		t.Fatal(err)
 	}
@@ -374,9 +376,9 @@ func TestRaft_Snapshot_Restart(t *testing.T) {
 
 /*
 func TestRaft_Snapshot_ErrorRecovery(t *testing.T) {
-	raft1, dir := getRaft(t, true, false)
-	raft2, dir2 := getRaft(t, false, false)
-	raft3, dir3 := getRaft(t, false, false)
+	raft1, dir := GetRaft(t, true, false)
+	raft2, dir2 := GetRaft(t, false, false)
+	raft3, dir3 := GetRaft(t, false, false)
 	defer os.RemoveAll(dir)
 	defer os.RemoveAll(dir2)
 	defer os.RemoveAll(dir3)
@@ -456,9 +458,9 @@ func TestRaft_Snapshot_ErrorRecovery(t *testing.T) {
 }*/
 
 func TestRaft_Snapshot_Take_Restore(t *testing.T) {
-	raft1, dir := getRaft(t, true, false)
+	raft1, dir := GetRaft(t, true, false)
 	defer os.RemoveAll(dir)
-	raft2, dir2 := getRaft(t, false, false)
+	raft2, dir2 := GetRaft(t, false, false)
 	defer os.RemoveAll(dir2)
 
 	addPeer(t, raft1, raft2)
@@ -590,7 +592,7 @@ func TestBoltSnapshotStore_Listing(t *testing.T) {
 		Level: hclog.Trace,
 	})
 
-	fsm, err := NewFSM(parent, logger)
+	fsm, err := NewFSM(parent, "", logger)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -655,7 +657,7 @@ func TestBoltSnapshotStore_CreateInstallSnapshot(t *testing.T) {
 		Level: hclog.Trace,
 	})
 
-	fsm, err := NewFSM(parent, logger)
+	fsm, err := NewFSM(parent, "", logger)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -753,7 +755,7 @@ func TestBoltSnapshotStore_CreateInstallSnapshot(t *testing.T) {
 		t.Fatal("expected snapshot installer object")
 	}
 
-	newFSM, err := NewFSM(filepath.Dir(installer.Filename()), logger)
+	newFSM, err := NewFSM(filepath.Dir(installer.Filename()), "", logger)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -812,7 +814,7 @@ func TestBoltSnapshotStore_CreateInstallSnapshot(t *testing.T) {
 
 		// Close/Reopen the db and make sure we still match
 		fsm.Close()
-		fsm, err = NewFSM(parent, logger)
+		fsm, err = NewFSM(parent, "", logger)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -892,7 +894,7 @@ func TestBoltSnapshotStore_BadPerm(t *testing.T) {
 	if err != nil {
 		t.Fatalf("err: %s", err)
 	}
-	if err = os.Chmod(dir2, 000); err != nil {
+	if err = os.Chmod(dir2, 0o00); err != nil {
 		t.Fatalf("err: %s", err)
 	}
 	defer os.Chmod(dir2, 777) // Set perms back for delete
