@@ -1,7 +1,11 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: MPL-2.0
+
 package framework
 
 import (
 	"context"
+	"fmt"
 	"net/http"
 	"reflect"
 	"strings"
@@ -9,8 +13,11 @@ import (
 	"testing"
 	"time"
 
+	"github.com/hashicorp/go-secure-stdlib/strutil"
+
 	"github.com/hashicorp/vault/sdk/helper/consts"
 	"github.com/hashicorp/vault/sdk/logical"
+	"github.com/stretchr/testify/require"
 )
 
 func BenchmarkBackendRoute(b *testing.B) {
@@ -45,6 +52,52 @@ func TestBackend_impl(t *testing.T) {
 	var _ logical.Backend = new(Backend)
 }
 
+func TestBackendHandleRequestFieldWarnings(t *testing.T) {
+	handler := func(ctx context.Context, req *logical.Request, data *FieldData) (*logical.Response, error) {
+		return &logical.Response{
+			Data: map[string]interface{}{
+				"an_int":   data.Get("an_int"),
+				"a_string": data.Get("a_string"),
+				"name":     data.Get("name"),
+			},
+		}, nil
+	}
+
+	backend := &Backend{
+		Paths: []*Path{
+			{
+				Pattern: "foo/bar/(?P<name>.+)",
+				Fields: map[string]*FieldSchema{
+					"an_int":   {Type: TypeInt},
+					"a_string": {Type: TypeString},
+					"name":     {Type: TypeString},
+				},
+				Operations: map[logical.Operation]OperationHandler{
+					logical.UpdateOperation: &PathOperation{Callback: handler},
+				},
+			},
+		},
+	}
+	ctx := context.Background()
+	resp, err := backend.HandleRequest(ctx, &logical.Request{
+		Operation: logical.UpdateOperation,
+		Path:      "foo/bar/baz",
+		Data: map[string]interface{}{
+			"an_int":        10,
+			"a_string":      "accepted",
+			"unrecognized1": "unrecognized",
+			"unrecognized2": 20.2,
+			"name":          "noop",
+		},
+	})
+	require.NoError(t, err)
+	require.NotNil(t, resp)
+	t.Log(resp.Warnings)
+	require.Len(t, resp.Warnings, 2)
+	require.True(t, strutil.StrListContains(resp.Warnings, "Endpoint ignored these unrecognized parameters: [unrecognized1 unrecognized2]"))
+	require.True(t, strutil.StrListContains(resp.Warnings, "Endpoint replaced the value of these parameters with the values captured from the endpoint's path: [name]"))
+}
+
 func TestBackendHandleRequest(t *testing.T) {
 	callback := func(ctx context.Context, req *logical.Request, data *FieldData) (*logical.Response, error) {
 		return &logical.Response{
@@ -66,7 +119,7 @@ func TestBackendHandleRequest(t *testing.T) {
 			{
 				Pattern: "foo/bar",
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeInt},
+					"value": {Type: TypeInt},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.ReadOperation: callback,
@@ -75,7 +128,7 @@ func TestBackendHandleRequest(t *testing.T) {
 			{
 				Pattern: "foo/baz/handler",
 				Fields: map[string]*FieldSchema{
-					"amount": &FieldSchema{Type: TypeInt},
+					"amount": {Type: TypeInt},
 				},
 				Operations: map[logical.Operation]OperationHandler{
 					logical.ReadOperation: &PathOperation{Callback: handler},
@@ -84,7 +137,7 @@ func TestBackendHandleRequest(t *testing.T) {
 			{
 				Pattern: "foo/both/handler",
 				Fields: map[string]*FieldSchema{
-					"amount": &FieldSchema{Type: TypeInt},
+					"amount": {Type: TypeInt},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.ReadOperation: callback,
@@ -228,10 +281,10 @@ func TestBackendHandleRequest_badwrite(t *testing.T) {
 
 	b := &Backend{
 		Paths: []*Path{
-			&Path{
+			{
 				Pattern: "foo/bar",
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeBool},
+					"value": {Type: TypeBool},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.UpdateOperation: callback,
@@ -240,16 +293,18 @@ func TestBackendHandleRequest_badwrite(t *testing.T) {
 		},
 	}
 
-	_, err := b.HandleRequest(context.Background(), &logical.Request{
+	resp, err := b.HandleRequest(context.Background(), &logical.Request{
 		Operation: logical.UpdateOperation,
 		Path:      "foo/bar",
 		Data:      map[string]interface{}{"value": "3false3"},
 	})
-
-	if err == nil {
-		t.Fatalf("should have thrown a conversion error")
+	if err != nil {
+		t.Fatalf("err: %s", err)
 	}
 
+	if !strings.Contains(resp.Data["error"].(string), "Field validation failed") {
+		t.Fatalf("bad: %#v", resp)
+	}
 }
 
 func TestBackendHandleRequest_404(t *testing.T) {
@@ -263,10 +318,10 @@ func TestBackendHandleRequest_404(t *testing.T) {
 
 	b := &Backend{
 		Paths: []*Path{
-			&Path{
+			{
 				Pattern: `foo/bar`,
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeInt},
+					"value": {Type: TypeInt},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.ReadOperation: callback,
@@ -288,10 +343,10 @@ func TestBackendHandleRequest_404(t *testing.T) {
 func TestBackendHandleRequest_help(t *testing.T) {
 	b := &Backend{
 		Paths: []*Path{
-			&Path{
+			{
 				Pattern: "foo/bar",
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeInt},
+					"value": {Type: TypeInt},
 				},
 				HelpSynopsis:    "foo",
 				HelpDescription: "bar",
@@ -360,6 +415,7 @@ func TestBackendHandleRequest_renewAuthCallback(t *testing.T) {
 		t.Fatalf("bad: %#v", v)
 	}
 }
+
 func TestBackendHandleRequest_renew(t *testing.T) {
 	called := new(uint32)
 	callback := func(context.Context, *logical.Request, *FieldData) (*logical.Response, error) {
@@ -485,10 +541,10 @@ func TestBackendHandleRequest_unsupportedOperation(t *testing.T) {
 
 	b := &Backend{
 		Paths: []*Path{
-			&Path{
+			{
 				Pattern: `foo/bar`,
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeInt},
+					"value": {Type: TypeInt},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.ReadOperation: callback,
@@ -518,10 +574,10 @@ func TestBackendHandleRequest_urlPriority(t *testing.T) {
 
 	b := &Backend{
 		Paths: []*Path{
-			&Path{
+			{
 				Pattern: `foo/(?P<value>\d+)`,
 				Fields: map[string]*FieldSchema{
-					"value": &FieldSchema{Type: TypeInt},
+					"value": {Type: TypeInt},
 				},
 				Callbacks: map[logical.Operation]OperationFunc{
 					logical.ReadOperation: callback,
@@ -613,13 +669,13 @@ func TestBackendSecret(t *testing.T) {
 		Match   bool
 	}{
 		"no match": {
-			[]*Secret{&Secret{Type: "foo"}},
+			[]*Secret{{Type: "foo"}},
 			"bar",
 			false,
 		},
 
 		"match": {
-			[]*Secret{&Secret{Type: "foo"}},
+			[]*Secret{{Type: "foo"}},
 			"foo",
 			true,
 		},
@@ -747,7 +803,6 @@ func TestFieldSchemaDefaultOrZero(t *testing.T) {
 }
 
 func TestInitializeBackend(t *testing.T) {
-
 	var inited bool
 	backend := &Backend{InitializeFunc: func(context.Context, *logical.InitializationRequest) error {
 		inited = true
@@ -760,3 +815,23 @@ func TestInitializeBackend(t *testing.T) {
 		t.Fatal("backend should be open")
 	}
 }
+
+// TestFieldTypeMethods tries to ensure our switch-case statements for the
+// FieldType "enum" are complete.
+func TestFieldTypeMethods(t *testing.T) {
+	unknownFormat := convertType(TypeInvalid).format
+
+	for i := TypeInvalid + 1; i < typeInvalidMax; i++ {
+		t.Run(fmt.Sprintf("%d", i), func(t *testing.T) {
+			if i.String() == TypeInvalid.String() {
+				t.Errorf("unknown type string for %d", i)
+			}
+
+			if convertType(i).format == unknownFormat {
+				t.Errorf("unknown schema for %d", i)
+			}
+
+			_ = i.Zero()
+		})
+	}
+}
