@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import { assert } from '@ember/debug';
 import Component from '@ember/component';
@@ -25,16 +30,16 @@ export default Component.extend({
 
   actions: {
     performTransaction() {
-      let args = [...arguments];
-      let messageArgs = this.messageArgs(...args);
+      const args = [...arguments];
+      const messageArgs = this.messageArgs(...args);
       return this.transaction(...args)
         .then(() => {
-          this.get('onSuccess')();
-          this.get('flashMessages').success(this.successMessage(...messageArgs));
+          this.onSuccess();
+          this.flashMessages.success(this.successMessage(...messageArgs));
         })
-        .catch(e => {
+        .catch((e) => {
           this.onError(...messageArgs);
-          this.get('flashMessages').success(this.errorMessage(e, ...messageArgs));
+          this.flashMessages.success(this.errorMessage(e, ...messageArgs));
         });
     },
   },
