@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Component from '@ember/component';
 import { task } from 'ember-concurrency';
 import { getOwner } from '@ember/application';
@@ -12,16 +17,16 @@ export default Component.extend({
   flashMessages: service(),
   isUploading: alias('restore.isRunning'),
   abortController: null,
-  restore: task(function*() {
+  restore: task(function* () {
     this.set('errors', null);
-    let adapter = getOwner(this).lookup('adapter:application');
+    const adapter = getOwner(this).lookup('adapter:application');
     try {
       let url = '/v1/sys/storage/raft/snapshot';
       if (this.forceRestore) {
         url = `${url}-force`;
       }
-      let file = new Blob([this.file], { type: 'application/gzip' });
-      let controller = new AbortController();
+      const file = new Blob([this.file], { type: 'application/gzip' });
+      const controller = new AbortController();
       this.set('abortController', controller);
       yield adapter.rawRequest(url, 'POST', { body: file, signal: controller.signal });
       this.flashMessages.success('The snapshot was successfully uploaded!');
@@ -33,7 +38,7 @@ export default Component.extend({
       if (e.json) {
         resp = yield e.json();
       }
-      let err = resp ? resp.errors : [e];
+      const err = resp ? resp.errors : [e];
       this.set('errors', err);
     }
   }),
