@@ -1,3 +1,9 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+import Component from '@glimmer/component';
 /**
  * @module ToolbarSecretLink
  * `ToolbarSecretLink` styles SecretLink for the Toolbar.
@@ -16,18 +22,8 @@
  *
  * @param type="" {String} - Use "add" to change icon
  */
-
-import OuterHTML from './outer-html';
-import { computed } from '@ember/object';
-
-export default OuterHTML.extend({
-  glyph: computed('type', function() {
-    if (this.type == 'add') {
-      return 'plus-plain';
-    } else {
-      return 'chevron-right';
-    }
-  }),
-  tagName: '',
-  supportsDataTestProperties: true,
-});
+export default class ToolbarSecretLink extends Component {
+  get glyph() {
+    return this.args.type === 'add' ? 'plus' : 'chevron-right';
+  }
+}
