@@ -1,6 +1,11 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import Controller from '@ember/controller';
-import { set, get } from '@ember/object';
+import { set } from '@ember/object';
 
 export default Controller.extend({
   store: service(),
@@ -19,23 +24,23 @@ export default Controller.extend({
       const hasErrors = codemirror.state.lint.marked.length > 0;
 
       if (!hasErrors) {
-        set(this.get('model'), attr, JSON.parse(val));
+        set(this.model, attr, JSON.parse(val));
       }
     },
 
     updateTtl(path, val) {
-      const model = this.get('model');
-      let valueToSet = val.enabled === true ? `${val.seconds}s` : undefined;
+      const model = this.model;
+      const valueToSet = val.enabled === true ? `${val.seconds}s` : undefined;
       set(model, path, valueToSet);
     },
 
     newModel() {
-      const model = this.get('model');
+      const model = this.model;
       const roleModel = model.get('role');
       model.unloadRecord();
-      const newModel = this.get('store').createRecord('ssh-sign', {
+      const newModel = this.store.createRecord('ssh-sign', {
         role: roleModel,
-        id: `${get(roleModel, 'backend')}-${get(roleModel, 'name')}`,
+        id: `${roleModel.backend}-${roleModel.name}`,
       });
       this.set('model', newModel);
     },
