@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import Controller from '@ember/controller';
 
@@ -11,7 +16,7 @@ export default Controller.extend({
         .seal()
         .then(() => {
           this.model.cluster.get('leaderNode').set('sealed', true);
-          this.get('auth').deleteCurrentToken();
+          this.auth.deleteCurrentToken();
           return this.transitionToRoute('vault.cluster.unseal');
         });
     },
