@@ -1,12 +1,21 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { run } from '@ember/runloop';
 import Helper from '@ember/component/helper';
-import { get } from '@ember/object';
+import Ember from 'ember';
 
 export default Helper.extend({
   disableInterval: false,
 
   compute(value, { interval }) {
-    if (get(this, 'disableInterval')) {
+    if (Ember.testing) {
+      // issues with flaky test, suspect it has to the do with the run loop not being cleared as intended farther down.
+      return;
+    }
+    if (this.disableInterval) {
       return;
     }
 
