@@ -1,4 +1,9 @@
-let cache = {};
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+const cache = {};
 
 export default {
   getItem(key) {
