@@ -1,11 +1,14 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+import { attr } from '@ember-data/model';
 import { computed } from '@ember/object';
-import DS from 'ember-data';
 import AuthConfig from '../auth-config';
 import { combineFieldGroups } from 'vault/utils/openapi-to-attrs';
 import fieldToAttrs from 'vault/utils/field-to-attrs';
 
-const { attr } = DS;
-
 export default AuthConfig.extend({
   useOpenAPI: true,
   // We have to leave this here because the backend doesn't support the file type yet.
@@ -15,7 +18,7 @@ export default AuthConfig.extend({
 
   googleCertsEndpoint: attr('string'),
 
-  fieldGroups: computed(function() {
+  fieldGroups: computed('newFields', function () {
     let groups = [
       { default: ['credentials'] },
       {
