@@ -1,11 +1,14 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+import { attr } from '@ember-data/model';
 import { computed } from '@ember/object';
-import DS from 'ember-data';
 import AuthConfig from '../auth-config';
 import fieldToAttrs from 'vault/utils/field-to-attrs';
 import { combineFieldGroups } from 'vault/utils/openapi-to-attrs';
 
-const { attr } = DS;
-
 export default AuthConfig.extend({
   useOpenAPI: true,
   orgName: attr('string', {
@@ -25,7 +28,7 @@ export default AuthConfig.extend({
       "Useful if using Vault's built-in MFA mechanisms. Will also cause certain other statuses to be ignored, such as PASSWORD_EXPIRED",
   }),
 
-  fieldGroups: computed(function() {
+  fieldGroups: computed('newFields', function () {
     let groups = [
       {
         default: ['orgName'],
