@@ -1,20 +1,21 @@
-import DS from 'ember-data';
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+import Model, { attr } from '@ember-data/model';
 import { computed } from '@ember/object';
-import { apiPath } from 'vault/macros/lazy-capabilities';
-import attachCapabilities from 'vault/lib/attach-capabilities';
+import lazyCapabilities, { apiPath } from 'vault/macros/lazy-capabilities';
 import { expandAttributeMeta } from 'vault/utils/field-to-attrs';
 
-const { attr } = DS;
-
-const Model = DS.Model.extend({
+export default Model.extend({
   idPrefix: 'alphabet/',
-  idForNav: computed('id', 'idPrefix', function() {
-    let modelId = this.id || '';
+  idForNav: computed('id', 'idPrefix', function () {
+    const modelId = this.id || '';
     return `${this.idPrefix}${modelId}`;
   }),
 
   name: attr('string', {
-    fieldValue: 'id',
     readOnly: true,
     subText: 'The alphabet name. Keep in mind that spaces are not allowed and this cannot be edited later.',
   }),
@@ -24,14 +25,11 @@ const Model = DS.Model.extend({
       'Provide the set of valid UTF-8 characters contained within both the input and transformed value. Read more.',
   }),
 
-  attrs: computed(function() {
-    let keys = ['name', 'alphabet'];
+  attrs: computed(function () {
+    const keys = ['name', 'alphabet'];
     return expandAttributeMeta(this, keys);
   }),
 
   backend: attr('string', { readOnly: true }),
-});
-
-export default attachCapabilities(Model, {
-  updatePath: apiPath`${'backend'}/alphabet/${'id'}`,
+  updatePath: lazyCapabilities(apiPath`${'backend'}/alphabet/${'id'}`, 'backend', 'id'),
 });
