@@ -1,9 +1,16 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Route from '@ember/routing/route';
-import UnloadModelRoute from 'vault/mixins/unload-model-route';
 import UnsavedModelRoute from 'vault/mixins/unsaved-model-route';
 import { singularize } from 'ember-inflector';
+import { inject as service } from '@ember/service';
+
+export default Route.extend(UnsavedModelRoute, {
+  store: service(),
 
-export default Route.extend(UnloadModelRoute, UnsavedModelRoute, {
   model() {
     const { item_type: itemType } = this.paramsFor('vault.cluster.access.method.item');
     const methodModel = this.modelFor('vault.cluster.access.method');
