@@ -1,9 +1,16 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Route from '@ember/routing/route';
 import ClusterRoute from 'vault/mixins/cluster-route';
 import { inject as service } from '@ember/service';
 
 export default Route.extend(ClusterRoute, {
+  store: service(),
   version: service(),
+
   beforeModel() {
     if (this.version.isOSS) {
       this.transitionTo('vault.cluster');
@@ -13,10 +20,4 @@ export default Route.extend(ClusterRoute, {
   model() {
     return this.store.queryRecord('license', {});
   },
-
-  actions: {
-    doRefresh() {
-      this.refresh();
-    },
-  },
 });
