@@ -1,18 +1,26 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { hash } from 'rsvp';
 import Route from '@ember/routing/route';
 import UnloadModelRoute from 'vault/mixins/unload-model-route';
+import { inject as service } from '@ember/service';
 
 export default Route.extend(UnloadModelRoute, {
+  store: service(),
+
   beforeModel() {
     const params = this.paramsFor(this.routeName);
-    let policyType = this.policyType();
+    const policyType = this.policyType();
     if (policyType === 'acl' && params.policy_name === 'root') {
       return this.transitionTo('vault.cluster.policies', 'acl');
     }
   },
 
   model(params) {
-    let type = this.policyType();
+    const type = this.policyType();
     return hash({
       policy: this.store.findRecord(`policy/${type}`, params.policy_name),
       capabilities: this.store.findRecord('capabilities', `sys/policies/${type}/${params.policy_name}`),
