@@ -1,17 +1,24 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import Route from '@ember/routing/route';
 export default Route.extend({
+  store: service(),
   flashMessages: service(),
   secretMountPath: service(),
   oldModel: null,
+
   model(params) {
-    let { backend } = params;
+    const { backend } = params;
     this.secretMountPath.update(backend);
     return this.store
       .query('secret-engine', {
         path: backend,
       })
-      .then(model => {
+      .then((model) => {
         if (model) {
           return model.get('firstObject');
         }
@@ -19,7 +26,7 @@ export default Route.extend({
   },
 
   afterModel(model, transition) {
-    let path = model && model.get('path');
+    const path = model && model.get('path');
     if (transition.targetName === this.routeName) {
       return this.replaceWith('vault.cluster.secrets.backend.list-root', path);
     }
