@@ -1,15 +1,22 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Route from '@ember/routing/route';
-import utils from 'vault/lib/key-utils';
 import UnloadModelRoute from 'vault/mixins/unload-model-route';
 import { normalizePath } from 'vault/utils/path-encoding-helpers';
+import { inject as service } from '@ember/service';
+import { parentKeyForKey } from 'core/utils/key-utils';
 
 export default Route.extend(UnloadModelRoute, {
+  store: service(),
   templateName: 'vault/cluster/secrets/backend/versions',
 
   beforeModel() {
-    let backendModel = this.modelFor('vault.cluster.secrets.backend');
+    const backendModel = this.modelFor('vault.cluster.secrets.backend');
     const { secret } = this.paramsFor(this.routeName);
-    const parentKey = utils.parentKeyForKey(secret);
+    const parentKey = parentKeyForKey(secret);
     if (backendModel.get('isV2KV')) {
       return;
     }
@@ -21,9 +28,9 @@ export default Route.extend(UnloadModelRoute, {
   },
 
   model(params) {
-    let { secret } = params;
+    const { secret } = params;
     const { backend } = this.paramsFor('vault.cluster.secrets.backend');
-    let id = normalizePath(secret);
+    const id = normalizePath(secret);
     return this.store.queryRecord('secret-v2', { id, backend });
   },
 });
