@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { get } from '@ember/object';
 import { assign } from '@ember/polyfills';
 import ApplicationSerializer from './application';
@@ -5,7 +10,7 @@ import ApplicationSerializer from './application';
 export default ApplicationSerializer.extend({
   secretDataPath: 'data.data',
   normalizeItems(payload) {
-    let path = this.secretDataPath;
+    const path = this.secretDataPath;
     // move response that is the contents of the secret from the dataPath
     // to `secret_data` so it will be `secretData` in the model
     payload.secret_data = get(payload, path);
@@ -17,7 +22,7 @@ export default ApplicationSerializer.extend({
     return payload;
   },
   serialize(snapshot) {
-    let secret = snapshot.belongsTo('secret');
+    const secret = snapshot.belongsTo('secret');
     // if both models failed to read from the server, we need to write without CAS
     if (secret.record.failedServerRead && snapshot.record.failedServerRead) {
       return {
