@@ -1,5 +1,9 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Service from '@ember/service';
-import { computed } from '@ember/object';
 
 const FLAGS = {
   vaultCloudNamespace: 'VAULT_CLOUD_ADMIN_NAMESPACE',
@@ -11,11 +15,10 @@ export default Service.extend({
     this.set('featureFlags', flags);
   },
 
-  managedNamespaceRoot: computed('featureFlags', function() {
-    const flags = this.featureFlags;
-    if (flags && flags.includes(FLAGS.vaultCloudNamespace)) {
+  get managedNamespaceRoot() {
+    if (this.featureFlags && this.featureFlags.includes(FLAGS.vaultCloudNamespace)) {
       return 'admin';
     }
     return null;
-  }),
+  },
 });
