@@ -1,17 +1,22 @@
-import DS from 'ember-data';
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+import Transform from '@ember-data/serializer/transform';
 import { typeOf } from '@ember/utils';
 /*
   DS.attr('object')
 */
-export default DS.Transform.extend({
-  deserialize: function(value) {
+export default Transform.extend({
+  deserialize: function (value) {
     if (typeOf(value) !== 'object') {
       return {};
     } else {
       return value;
     }
   },
-  serialize: function(value) {
+  serialize: function (value) {
     if (typeOf(value) !== 'object') {
       return {};
     } else {
