@@ -1,3 +1,10 @@
+import Component from '@glimmer/component';
+
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 /**
  * @module BoxRadio
  * BoxRadio components are used to display options for a radio selection.
@@ -16,11 +23,8 @@
  * @param {string} [tooltipMessage=default] - The message that shows in the tooltip if the radio option is disabled
  */
 
-import Component from '@ember/component';
-import layout from '../templates/components/box-radio';
-
-export default Component.extend({
-  layout,
-  disabled: false,
-  tooltipMessage: 'This option is not available to you at this time.',
-});
+export default class BoxRadio extends Component {
+  get tooltipMessage() {
+    return this.args.tooltipMessage || 'This option is not available to you at this time.';
+  }
+}
