@@ -1,3 +1,9 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
+/* eslint-disable ember/no-observers */
 import { inject as service } from '@ember/service';
 import { assert } from '@ember/debug';
 import Helper from '@ember/component/helper';
@@ -14,6 +20,7 @@ const POSSIBLE_FEATURES = [
   'Namespaces',
   'KMIP',
   'Transform Secrets Engine',
+  'Key Management Secrets Engine',
 ];
 
 export function hasFeature(featureName, features) {
@@ -26,11 +33,10 @@ export function hasFeature(featureName, features) {
 
 export default Helper.extend({
   version: service(),
-  /* eslint-disable-next-line ember/no-observers */
-  onFeaturesChange: observer('version.features.[]', function() {
+  onFeaturesChange: observer('version.features.[]', function () {
     this.recompute();
   }),
   compute([featureName]) {
-    return hasFeature(featureName, this.get('version.features'));
+    return hasFeature(featureName, this.version.features);
   },
 });
