@@ -1,10 +1,16 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { helper as buildHelper } from '@ember/component/helper';
+import { assert } from '@ember/debug';
 
 export const MESSAGE_TYPES = {
   info: {
     class: 'is-info',
     glyphClass: 'has-text-info',
-    glyph: 'info-circle-fill',
+    glyph: 'info',
     text: 'Info',
   },
   success: {
@@ -16,13 +22,13 @@ export const MESSAGE_TYPES = {
   danger: {
     class: 'is-danger',
     glyphClass: 'has-text-danger',
-    glyph: 'cancel-square-fill',
+    glyph: 'x-square-fill',
     text: 'Error',
   },
   warning: {
     class: 'is-highlight',
     glyphClass: 'has-text-highlight',
-    glyph: 'alert-triangle',
+    glyph: 'alert-triangle-fill',
     text: 'Warning',
   },
   loading: {
@@ -31,9 +37,17 @@ export const MESSAGE_TYPES = {
     glyph: 'loading',
     text: 'Loading',
   },
+  rotation: {
+    class: 'is-info',
+    glyphClass: 'has-text-grey',
+    glyph: 'rotate-cw',
+  },
 };
 
 export function messageTypes([type]) {
+  if (!(type in MESSAGE_TYPES)) {
+    assert('type is not a valid message type.');
+  }
   return MESSAGE_TYPES[type];
 }
 
