@@ -1,14 +1,19 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import Route from '@ember/routing/route';
 import { inject as service } from '@ember/service';
 
-export default Route.extend({
-  store: service(),
-  secretMountPath: service(),
-  pathHelp: service(),
+export default class KmipRoleRoute extends Route {
+  @service store;
+  @service secretMountPath;
+  @service pathHelp;
 
   beforeModel() {
     return this.pathHelp.getNewModel('kmip/role', this.secretMountPath.currentPath);
-  },
+  }
 
   model(params) {
     return this.store.queryRecord('kmip/role', {
@@ -16,11 +21,11 @@ export default Route.extend({
       scope: params.scope_name,
       id: params.role_name,
     });
-  },
+  }
 
   setupController(controller) {
-    this._super(...arguments);
-    let { scope_name: scope, role_name: role } = this.paramsFor('role');
+    super.setupController(...arguments);
+    const { scope_name: scope, role_name: role } = this.paramsFor('role');
     controller.setProperties({ role, scope });
-  },
-});
+  }
+}
