@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import { setProperties } from '@ember/object';
 import { hash } from 'rsvp';
@@ -10,26 +15,23 @@ export default Route.extend(ClusterRoute, {
   auth: service(),
 
   beforeModel() {
-    return this.get('version')
-      .fetchFeatures()
-      .then(() => {
-        return this._super(...arguments);
-      });
+    return this.version.fetchFeatures().then(() => {
+      return this._super(...arguments);
+    });
   },
 
   model() {
-    const activeClusterId = this.get('auth.activeCluster');
-    return this.store.peekRecord('cluster', activeClusterId);
+    return this.auth.activeCluster;
   },
 
   afterModel(model) {
     return hash({
       canEnablePrimary: this.store
         .findRecord('capabilities', 'sys/replication/primary/enable')
-        .then(c => c.get('canUpdate')),
+        .then((c) => c.get('canUpdate')),
       canEnableSecondary: this.store
         .findRecord('capabilities', 'sys/replication/secondary/enable')
-        .then(c => c.get('canUpdate')),
+        .then((c) => c.get('canUpdate')),
     }).then(({ canEnablePrimary, canEnableSecondary }) => {
       setProperties(model, {
         canEnablePrimary,
