@@ -1,10 +1,15 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { inject as service } from '@ember/service';
 import Route from '@ember/routing/route';
 
 export default Route.extend({
   replicationMode: service(),
   beforeModel() {
-    this.get('replicationMode').setMode(null);
+    this.replicationMode.setMode(null);
   },
   model() {
     return this.modelFor('application');
