@@ -1,3 +1,8 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { Base } from '../credentials';
 import { clickable, value, create, fillable, isPresent } from 'ember-cli-page-object';
 
@@ -11,9 +16,7 @@ export default create({
   commonNameValue: value('[data-test-input="commonName"]'),
   submit: clickable('[data-test-secret-generate]'),
   back: clickable('[data-test-secret-generate-back]'),
-  generateOTP: async function() {
-    await this.user('admin')
-      .ip('192.168.1.1')
-      .submit();
+  generateOTP: async function () {
+    await this.user('admin').ip('192.168.1.1').submit();
   },
 });
