@@ -1,11 +1,15 @@
+/**
+ * Copyright (c) HashiCorp, Inc.
+ * SPDX-License-Identifier: BUSL-1.1
+ */
+
 import { create, visitable, collection, clickable, text } from 'ember-cli-page-object';
 import uiPanel from 'vault/tests/pages/components/console/ui-panel';
 
 export default create({
-  console: uiPanel,
   consoleToggle: clickable('[data-test-console-toggle]'),
   visit: visitable('/vault/secrets'),
-  rows: collection('[data-test-secret-backend-row]', {
+  rows: collection('[data-test-secrets-backend-link]', {
     path: text('[data-test-secret-path]'),
     menu: clickable('[data-test-popup-menu-trigger]'),
   }),
@@ -18,4 +22,5 @@ export default create({
   confirmDisable: clickable('[data-test-confirm-button]', {
     testContainer: '#ember-testing',
   }),
+  console: uiPanel,
 });
