@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package vault
 
 import (
@@ -10,14 +13,15 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"math"
+	"strconv"
 	"strings"
 	"sync"
 	"time"
 
 	"github.com/armon/go-metrics"
-	"github.com/hashicorp/errwrap"
+	"github.com/hashicorp/go-secure-stdlib/strutil"
 	"github.com/hashicorp/vault/sdk/helper/jsonutil"
-	"github.com/hashicorp/vault/sdk/helper/strutil"
 	"github.com/hashicorp/vault/sdk/logical"
 	"github.com/hashicorp/vault/sdk/physical"
 	"go.uber.org/atomic"
@@ -30,6 +34,11 @@ const (
 
 	// termSize the number of bytes used for the key term.
 	termSize = 4
+
+	autoRotateCheckInterval = 5 * time.Minute
+	legacyRotateReason      = "legacy rotation"
+	// The keyring is persisted before the root key.
+	keyringTimeout = 1 * time.Second
 )
 
 // Versions of the AESGCM storage methodology
@@ -45,7 +54,11 @@ type barrierInit struct {
 }
 
 // Validate AESGCMBarrier satisfies SecurityBarrier interface
-var _ SecurityBarrier = &AESGCMBarrier{}
+var (
+	_                      SecurityBarrier = &AESGCMBarrier{}
+	barrierEncryptsMetric                  = []string{"barrier", "estimated_encryptions"}
+	barrierRotationsMetric                 = []string{"barrier", "auto_rotation"}
+)
 
 // AESGCMBarrier is a SecurityBarrier implementation that uses the AES
 // cipher core and the Galois Counter Mode block mode. It defaults to
@@ -73,6 +86,30 @@ type AESGCMBarrier struct {
 	currentAESGCMVersionByte byte
 
 	initialized atomic.Bool
+
+	UnaccountedEncryptions *atomic.Int64
+	// Used only for testing
+	RemoteEncryptions     *atomic.Int64
+	totalLocalEncryptions *atomic.Int64
+}
+
+func (b *AESGCMBarrier) RotationConfig() (kc KeyRotationConfig, err error) {
+	if b.keyring == nil {
+		return kc, errors.New("keyring not yet present")
+	}
+	return b.keyring.rotationConfig.Clone(), nil
+}
+
+func (b *AESGCMBarrier) SetRotationConfig(ctx context.Context, rotConfig KeyRotationConfig) error {
+	b.l.Lock()
+	defer b.l.Unlock()
+	rotConfig.Sanitize()
+	if !rotConfig.Equals(b.keyring.rotationConfig) {
+		b.keyring.rotationConfig = rotConfig
+
+		return b.persistKeyring(ctx, b.keyring)
+	}
+	return nil
 }
 
 // NewAESGCMBarrier is used to construct a new barrier that uses
@@ -83,12 +120,15 @@ func NewAESGCMBarrier(physical physical.Backend) (*AESGCMBarrier, error) {
 		sealed:                   true,
 		cache:                    make(map[uint32]cipher.AEAD),
 		currentAESGCMVersionByte: byte(AESGCMVersion2),
+		UnaccountedEncryptions:   atomic.NewInt64(0),
+		RemoteEncryptions:        atomic.NewInt64(0),
+		totalLocalEncryptions:    atomic.NewInt64(0),
 	}
 	return b, nil
 }
 
 // Initialized checks if the barrier has been initialized
-// and has a master key set.
+// and has a root key set.
 func (b *AESGCMBarrier) Initialized(ctx context.Context) (bool, error) {
 	if b.initialized.Load() {
 		return true, nil
@@ -97,7 +137,7 @@ func (b *AESGCMBarrier) Initialized(ctx context.Context) (bool, error) {
 	// Read the keyring file
 	keys, err := b.backend.List(ctx, keyringPrefix)
 	if err != nil {
-		return false, errwrap.Wrapf("failed to check for initialization: {{err}}", err)
+		return false, fmt.Errorf("failed to check for initialization: %w", err)
 	}
 	if strutil.StrListContains(keys, "keyring") {
 		b.initialized.Store(true)
@@ -107,15 +147,15 @@ func (b *AESGCMBarrier) Initialized(ctx context.Context) (bool, error) {
 	// Fallback, check for the old sentinel file
 	out, err := b.backend.Get(ctx, barrierInitPath)
 	if err != nil {
-		return false, errwrap.Wrapf("failed to check for initialization: {{err}}", err)
+		return false, fmt.Errorf("failed to check for initialization: %w", err)
 	}
 	b.initialized.Store(out != nil)
 	return out != nil, nil
 }
 
 // Initialize works only if the barrier has not been initialized
-// and makes use of the given master key.
-func (b *AESGCMBarrier) Initialize(ctx context.Context, key, sealKey []byte, reader io.Reader) error {
+// and makes use of the given root key.
+func (b *AESGCMBarrier) Initialize(ctx context.Context, key []byte, sealKey []byte, reader io.Reader) error {
 	// Verify the key size
 	min, max := b.KeyLength()
 	if len(key) < min || len(key) > max {
@@ -130,21 +170,21 @@ func (b *AESGCMBarrier) Initialize(ctx context.Context, key, sealKey []byte, rea
 	}
 
 	// Generate encryption key
-	encrypt, err := b.GenerateKey(reader)
+	encryptionKey, err := b.GenerateKey(reader)
 	if err != nil {
-		return errwrap.Wrapf("failed to generate encryption key: {{err}}", err)
+		return fmt.Errorf("failed to generate encryption key: %w", err)
 	}
 
 	// Create a new keyring, install the keys
 	keyring := NewKeyring()
-	keyring = keyring.SetMasterKey(key)
+	keyring = keyring.SetRootKey(key)
 	keyring, err = keyring.AddKey(&Key{
 		Term:    1,
 		Version: 1,
-		Value:   encrypt,
+		Value:   encryptionKey,
 	})
 	if err != nil {
-		return errwrap.Wrapf("failed to create keyring: {{err}}", err)
+		return fmt.Errorf("failed to create keyring: %w", err)
 	}
 
 	err = b.persistKeyring(ctx, keyring)
@@ -153,7 +193,7 @@ func (b *AESGCMBarrier) Initialize(ctx context.Context, key, sealKey []byte, rea
 	}
 
 	if len(sealKey) > 0 {
-		primary, err := b.aeadFromKey(encrypt)
+		primary, err := b.aeadFromKey(encryptionKey)
 		if err != nil {
 			return err
 		}
@@ -163,7 +203,7 @@ func (b *AESGCMBarrier) Initialize(ctx context.Context, key, sealKey []byte, rea
 			Value: sealKey,
 		})
 		if err != nil {
-			return errwrap.Wrapf("failed to store new seal key: {{err}}", err)
+			return fmt.Errorf("failed to store new seal key: %w", err)
 		}
 	}
 
@@ -171,25 +211,37 @@ func (b *AESGCMBarrier) Initialize(ctx context.Context, key, sealKey []byte, rea
 }
 
 // persistKeyring is used to write out the keyring using the
-// master key to encrypt it.
+// root key to encrypt it.
 func (b *AESGCMBarrier) persistKeyring(ctx context.Context, keyring *Keyring) error {
+	return b.persistKeyringInternal(ctx, keyring, false)
+}
+
+// persistKeyringBestEffort is like persistKeyring but 'best effort', ie times out early
+// for non critical keyring writes (encryption/rotation tracking)
+func (b *AESGCMBarrier) persistKeyringBestEffort(ctx context.Context, keyring *Keyring) error {
+	return b.persistKeyringInternal(ctx, keyring, true)
+}
+
+// persistKeyring is used to write out the keyring using the
+// root key to encrypt it.
+func (b *AESGCMBarrier) persistKeyringInternal(ctx context.Context, keyring *Keyring, bestEffort bool) error {
 	// Create the keyring entry
 	keyringBuf, err := keyring.Serialize()
 	defer memzero(keyringBuf)
 	if err != nil {
-		return errwrap.Wrapf("failed to serialize keyring: {{err}}", err)
+		return fmt.Errorf("failed to serialize keyring: %w", err)
 	}
 
 	// Create the AES-GCM
-	gcm, err := b.aeadFromKey(keyring.MasterKey())
+	gcm, err := b.aeadFromKey(keyring.RootKey())
 	if err != nil {
-		return err
+		return fmt.Errorf("failed to retrieve AES-GCM AEAD from root key: %w", err)
 	}
 
 	// Encrypt the barrier init value
 	value, err := b.encrypt(keyringPath, initialKeyTerm, gcm, keyringBuf)
 	if err != nil {
-		return err
+		return fmt.Errorf("failed to encrypt barrier initial value: %w", err)
 	}
 
 	// Create the keyring physical entry
@@ -197,40 +249,54 @@ func (b *AESGCMBarrier) persistKeyring(ctx context.Context, keyring *Keyring) er
 		Key:   keyringPath,
 		Value: value,
 	}
-	if err := b.backend.Put(ctx, pe); err != nil {
-		return errwrap.Wrapf("failed to persist keyring: {{err}}", err)
+
+	ctxKeyring := ctx
+
+	if bestEffort {
+		// We reduce the timeout on the initial 'put' but if this succeeds we will
+		// allow longer later on when we try to persist the root key .
+		var cancelKeyring func()
+		ctxKeyring, cancelKeyring = context.WithTimeout(ctx, keyringTimeout)
+		defer cancelKeyring()
 	}
 
-	// Serialize the master key value
+	if err := b.backend.Put(ctxKeyring, pe); err != nil {
+		return fmt.Errorf("failed to persist keyring: %w", err)
+	}
+
+	// Serialize the root key value
 	key := &Key{
 		Term:    1,
 		Version: 1,
-		Value:   keyring.MasterKey(),
+		Value:   keyring.RootKey(),
 	}
 	keyBuf, err := key.Serialize()
 	defer memzero(keyBuf)
 	if err != nil {
-		return errwrap.Wrapf("failed to serialize master key: {{err}}", err)
+		return fmt.Errorf("failed to serialize root key: %w", err)
 	}
 
-	// Encrypt the master key
+	// Encrypt the root key
 	activeKey := keyring.ActiveKey()
 	aead, err := b.aeadFromKey(activeKey.Value)
 	if err != nil {
-		return err
+		return fmt.Errorf("failed to retrieve AES-GCM AEAD from active key: %w", err)
 	}
-	value, err = b.encrypt(masterKeyPath, activeKey.Term, aead, keyBuf)
+	value, err = b.encryptTracked(rootKeyPath, activeKey.Term, aead, keyBuf)
 	if err != nil {
-		return err
+		return fmt.Errorf("failed to encrypt and track active key value: %w", err)
 	}
 
-	// Update the masterKeyPath for standby instances
+	// Update the rootKeyPath for standby instances
 	pe = &physical.Entry{
-		Key:   masterKeyPath,
+		Key:   rootKeyPath,
 		Value: value,
 	}
+
+	// Use the longer timeout from the original context, for the follow-up write
+	// to persist the root key, as the initial storage of the keyring was successful.
 	if err := b.backend.Put(ctx, pe); err != nil {
-		return errwrap.Wrapf("failed to persist master key: {{err}}", err)
+		return fmt.Errorf("failed to persist root key: %w", err)
 	}
 	return nil
 }
@@ -258,14 +324,14 @@ func (b *AESGCMBarrier) Sealed() (bool, error) {
 	return sealed, nil
 }
 
-// VerifyMaster is used to check if the given key matches the master key
-func (b *AESGCMBarrier) VerifyMaster(key []byte) error {
+// VerifyRoot is used to check if the given key matches the root key
+func (b *AESGCMBarrier) VerifyRoot(key []byte) error {
 	b.l.RLock()
 	defer b.l.RUnlock()
 	if b.sealed {
 		return ErrBarrierSealed
 	}
-	if subtle.ConstantTimeCompare(key, b.keyring.MasterKey()) != 1 {
+	if subtle.ConstantTimeCompare(key, b.keyring.RootKey()) != 1 {
 		return ErrBarrierInvalidKey
 	}
 	return nil
@@ -279,7 +345,7 @@ func (b *AESGCMBarrier) ReloadKeyring(ctx context.Context) error {
 	defer b.l.Unlock()
 
 	// Create the AES-GCM
-	gcm, err := b.aeadFromKey(b.keyring.MasterKey())
+	gcm, err := b.aeadFromKey(b.keyring.RootKey())
 	if err != nil {
 		return err
 	}
@@ -287,7 +353,7 @@ func (b *AESGCMBarrier) ReloadKeyring(ctx context.Context) error {
 	// Read in the keyring
 	out, err := b.backend.Get(ctx, keyringPath)
 	if err != nil {
-		return errwrap.Wrapf("failed to check for keyring: {{err}}", err)
+		return fmt.Errorf("failed to check for keyring: %w", err)
 	}
 
 	// Ensure that the keyring exists. This should never happen,
@@ -312,10 +378,19 @@ func (b *AESGCMBarrier) ReloadKeyring(ctx context.Context) error {
 		return err
 	}
 
-	// Recover the keyring
-	keyring, err := DeserializeKeyring(plain)
+	// Reset enc. counters, this may be a leadership change
+	b.totalLocalEncryptions.Store(0)
+	b.totalLocalEncryptions.Store(0)
+	b.UnaccountedEncryptions.Store(0)
+	b.RemoteEncryptions.Store(0)
+
+	return b.recoverKeyring(plain)
+}
+
+func (b *AESGCMBarrier) recoverKeyring(plaintext []byte) error {
+	keyring, err := DeserializeKeyring(plaintext)
 	if err != nil {
-		return errwrap.Wrapf("keyring deserialization failed: {{err}}", err)
+		return fmt.Errorf("keyring deserialization failed: %w", err)
 	}
 
 	// Setup the keyring and finish
@@ -324,19 +399,19 @@ func (b *AESGCMBarrier) ReloadKeyring(ctx context.Context) error {
 	return nil
 }
 
-// ReloadMasterKey is used to re-read the underlying masterkey.
-// This is used for HA deployments to ensure the latest master key
+// ReloadRootKey is used to re-read the underlying root key.
+// This is used for HA deployments to ensure the latest root key
 // is available for keyring reloading.
-func (b *AESGCMBarrier) ReloadMasterKey(ctx context.Context) error {
-	// Read the masterKeyPath upgrade
-	out, err := b.Get(ctx, masterKeyPath)
+func (b *AESGCMBarrier) ReloadRootKey(ctx context.Context) error {
+	// Read the rootKeyPath upgrade
+	out, err := b.Get(ctx, rootKeyPath)
 	if err != nil {
-		return errwrap.Wrapf("failed to read master key path: {{err}}", err)
+		return fmt.Errorf("failed to read root key path: %w", err)
 	}
 
-	// The masterKeyPath could be missing (backwards incompatible),
+	// The rootKeyPath could be missing (backwards incompatible),
 	// we can ignore this and attempt to make progress with the current
-	// master key.
+	// root key.
 	if out == nil {
 		return nil
 	}
@@ -345,35 +420,35 @@ func (b *AESGCMBarrier) ReloadMasterKey(ctx context.Context) error {
 	b.l.Lock()
 	defer b.l.Unlock()
 
-	out, err = b.lockSwitchedGet(ctx, masterKeyPath, false)
+	out, err = b.lockSwitchedGet(ctx, rootKeyPath, false)
 	if err != nil {
-		return errwrap.Wrapf("failed to read master key path: {{err}}", err)
+		return fmt.Errorf("failed to read root key path: %w", err)
 	}
 
 	if out == nil {
 		return nil
 	}
 
-	// Deserialize the master key
+	// Deserialize the root key
 	key, err := DeserializeKey(out.Value)
 	memzero(out.Value)
 	if err != nil {
-		return errwrap.Wrapf("failed to deserialize key: {{err}}", err)
+		return fmt.Errorf("failed to deserialize key: %w", err)
 	}
 
-	// Check if the master key is the same
-	if subtle.ConstantTimeCompare(b.keyring.MasterKey(), key.Value) == 1 {
+	// Check if the root key is the same
+	if subtle.ConstantTimeCompare(b.keyring.RootKey(), key.Value) == 1 {
 		return nil
 	}
 
-	// Update the master key
+	// Update the root key
 	oldKeyring := b.keyring
-	b.keyring = b.keyring.SetMasterKey(key.Value)
+	b.keyring = b.keyring.SetRootKey(key.Value)
 	oldKeyring.Zeroize(false)
 	return nil
 }
 
-// Unseal is used to provide the master key which permits the barrier
+// Unseal is used to provide the root key which permits the barrier
 // to be unsealed. If the key is not correct, the barrier remains sealed.
 func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 	b.l.Lock()
@@ -393,7 +468,7 @@ func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 	// Read in the keyring
 	out, err := b.backend.Get(ctx, keyringPath)
 	if err != nil {
-		return errwrap.Wrapf("failed to check for keyring: {{err}}", err)
+		return fmt.Errorf("failed to check for keyring: %w", err)
 	}
 	if out != nil {
 		// Verify the term is always just one
@@ -413,21 +488,20 @@ func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 		}
 
 		// Recover the keyring
-		keyring, err := DeserializeKeyring(plain)
+		err = b.recoverKeyring(plain)
 		if err != nil {
-			return errwrap.Wrapf("keyring deserialization failed: {{err}}", err)
+			return fmt.Errorf("keyring deserialization failed: %w", err)
 		}
 
-		// Setup the keyring and finish
-		b.keyring = keyring
 		b.sealed = false
+
 		return nil
 	}
 
 	// Read the barrier initialization key
 	out, err = b.backend.Get(ctx, barrierInitPath)
 	if err != nil {
-		return errwrap.Wrapf("failed to check for initialization: {{err}}", err)
+		return fmt.Errorf("failed to check for initialization: %w", err)
 	}
 	if out == nil {
 		return ErrBarrierNotInit
@@ -457,9 +531,9 @@ func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 
 	// Setup a new keyring, this is for backwards compatibility
 	keyringNew := NewKeyring()
-	keyring := keyringNew.SetMasterKey(key)
+	keyring := keyringNew.SetRootKey(key)
 
-	// AddKey reuses the master, so we are only zeroizing after this call
+	// AddKey reuses the root, so we are only zeroizing after this call
 	defer keyringNew.Zeroize(false)
 
 	keyring, err = keyring.AddKey(&Key{
@@ -468,7 +542,7 @@ func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 		Value:   init.Key,
 	})
 	if err != nil {
-		return errwrap.Wrapf("failed to create keyring: {{err}}", err)
+		return fmt.Errorf("failed to create keyring: %w", err)
 	}
 	if err := b.persistKeyring(ctx, keyring); err != nil {
 		return err
@@ -476,12 +550,13 @@ func (b *AESGCMBarrier) Unseal(ctx context.Context, key []byte) error {
 
 	// Delete the old barrier entry
 	if err := b.backend.Delete(ctx, barrierInitPath); err != nil {
-		return errwrap.Wrapf("failed to delete barrier init file: {{err}}", err)
+		return fmt.Errorf("failed to delete barrier init file: %w", err)
 	}
 
 	// Set the vault as unsealed
 	b.keyring = keyring
 	b.sealed = false
+
 	return nil
 }
 
@@ -501,7 +576,7 @@ func (b *AESGCMBarrier) Seal() error {
 
 // Rotate is used to create a new encryption key. All future writes
 // should use the new key, while old values should still be decryptable.
-func (b *AESGCMBarrier) Rotate(ctx context.Context, reader io.Reader) (uint32, error) {
+func (b *AESGCMBarrier) Rotate(ctx context.Context, randomSource io.Reader) (uint32, error) {
 	b.l.Lock()
 	defer b.l.Unlock()
 	if b.sealed {
@@ -509,9 +584,9 @@ func (b *AESGCMBarrier) Rotate(ctx context.Context, reader io.Reader) (uint32, e
 	}
 
 	// Generate a new key
-	encrypt, err := b.GenerateKey(reader)
+	encrypt, err := b.GenerateKey(randomSource)
 	if err != nil {
-		return 0, errwrap.Wrapf("failed to generate encryption key: {{err}}", err)
+		return 0, fmt.Errorf("failed to generate encryption key: %w", err)
 	}
 
 	// Get the next term
@@ -525,7 +600,7 @@ func (b *AESGCMBarrier) Rotate(ctx context.Context, reader io.Reader) (uint32, e
 		Value:   encrypt,
 	})
 	if err != nil {
-		return 0, errwrap.Wrapf("failed to add new encryption key: {{err}}", err)
+		return 0, fmt.Errorf("failed to add new encryption key: %w", err)
 	}
 
 	// Persist the new keyring
@@ -533,8 +608,14 @@ func (b *AESGCMBarrier) Rotate(ctx context.Context, reader io.Reader) (uint32, e
 		return 0, err
 	}
 
+	// Clear encryption tracking
+	b.RemoteEncryptions.Store(0)
+	b.totalLocalEncryptions.Store(0)
+	b.UnaccountedEncryptions.Store(0)
+
 	// Swap the keyrings
 	b.keyring = newKeyring
+
 	return newTerm, nil
 }
 
@@ -564,7 +645,7 @@ func (b *AESGCMBarrier) CreateUpgrade(ctx context.Context, term uint32) error {
 	}
 
 	key := fmt.Sprintf("%s%d", keyringUpgradePrefix, prevTerm)
-	value, err := b.encrypt(key, prevTerm, primary, buf)
+	value, err := b.encryptTracked(key, prevTerm, primary, buf)
 	b.l.RUnlock()
 	if err != nil {
 		return err
@@ -641,7 +722,7 @@ func (b *AESGCMBarrier) CheckUpgrade(ctx context.Context) (bool, uint32, error)
 	// Update the keyring
 	newKeyring, err := b.keyring.AddKey(key)
 	if err != nil {
-		return false, 0, errwrap.Wrapf("failed to add new encryption key: {{err}}", err)
+		return false, 0, fmt.Errorf("failed to add new encryption key: %w", err)
 	}
 	b.keyring = newKeyring
 
@@ -665,16 +746,17 @@ func (b *AESGCMBarrier) ActiveKeyInfo() (*KeyInfo, error) {
 	info := &KeyInfo{
 		Term:        int(term),
 		InstallTime: key.InstallTime,
+		Encryptions: b.encryptions(),
 	}
 	return info, nil
 }
 
-// Rekey is used to change the master key used to protect the keyring
+// Rekey is used to change the root key used to protect the keyring
 func (b *AESGCMBarrier) Rekey(ctx context.Context, key []byte) error {
 	b.l.Lock()
 	defer b.l.Unlock()
 
-	newKeyring, err := b.updateMasterKeyCommon(key)
+	newKeyring, err := b.updateRootKeyCommon(key)
 	if err != nil {
 		return err
 	}
@@ -691,13 +773,13 @@ func (b *AESGCMBarrier) Rekey(ctx context.Context, key []byte) error {
 	return nil
 }
 
-// SetMasterKey updates the keyring's in-memory master key but does not persist
+// SetRootKey updates the keyring's in-memory root key but does not persist
 // anything to storage
-func (b *AESGCMBarrier) SetMasterKey(key []byte) error {
+func (b *AESGCMBarrier) SetRootKey(key []byte) error {
 	b.l.Lock()
 	defer b.l.Unlock()
 
-	newKeyring, err := b.updateMasterKeyCommon(key)
+	newKeyring, err := b.updateRootKeyCommon(key)
 	if err != nil {
 		return err
 	}
@@ -709,9 +791,9 @@ func (b *AESGCMBarrier) SetMasterKey(key []byte) error {
 	return nil
 }
 
-// Performs common tasks related to updating the master key; note that the lock
+// Performs common tasks related to updating the root key; note that the lock
 // must be held before calling this function
-func (b *AESGCMBarrier) updateMasterKeyCommon(key []byte) (*Keyring, error) {
+func (b *AESGCMBarrier) updateRootKeyCommon(key []byte) (*Keyring, error) {
 	if b.sealed {
 		return nil, ErrBarrierSealed
 	}
@@ -722,7 +804,7 @@ func (b *AESGCMBarrier) updateMasterKeyCommon(key []byte) (*Keyring, error) {
 		return nil, fmt.Errorf("key size must be %d or %d", min, max)
 	}
 
-	return b.keyring.SetMasterKey(key), nil
+	return b.keyring.SetRootKey(key), nil
 }
 
 // Put is used to insert or update an entry
@@ -745,7 +827,7 @@ func (b *AESGCMBarrier) Put(ctx context.Context, entry *logical.StorageEntry) er
 }
 
 func (b *AESGCMBarrier) putInternal(ctx context.Context, term uint32, primary cipher.AEAD, entry *logical.StorageEntry) error {
-	value, err := b.encrypt(entry.Key, term, primary, entry.Value)
+	value, err := b.encryptTracked(entry.Key, term, primary, entry.Value)
 	if err != nil {
 		return err
 	}
@@ -815,7 +897,7 @@ func (b *AESGCMBarrier) lockSwitchedGet(ctx context.Context, key string, getLock
 	// Decrypt the ciphertext
 	plain, err := b.decrypt(key, gcm, pe.Value)
 	if err != nil {
-		return nil, errwrap.Wrapf("decryption failed: {{err}}", err)
+		return nil, fmt.Errorf("decryption failed: %w", err)
 	}
 
 	// Wrap in a logical entry
@@ -894,7 +976,7 @@ func (b *AESGCMBarrier) aeadFromKey(key []byte) (cipher.AEAD, error) {
 	// Create the AES cipher
 	aesCipher, err := aes.NewCipher(key)
 	if err != nil {
-		return nil, errwrap.Wrapf("failed to create cipher: {{err}}", err)
+		return nil, fmt.Errorf("failed to create cipher: %w", err)
 	}
 
 	// Create the GCM mode AEAD
@@ -909,10 +991,13 @@ func (b *AESGCMBarrier) aeadFromKey(key []byte) (cipher.AEAD, error) {
 func (b *AESGCMBarrier) encrypt(path string, term uint32, gcm cipher.AEAD, plain []byte) ([]byte, error) {
 	// Allocate the output buffer with room for tern, version byte,
 	// nonce, GCM tag and the plaintext
-	capacity := termSize + 1 + gcm.NonceSize() + gcm.Overhead() + len(plain)
-	if capacity < 0 {
+
+	extra := termSize + 1 + gcm.NonceSize() + gcm.Overhead()
+	if len(plain) > math.MaxInt-extra {
 		return nil, ErrPlaintextTooLarge
 	}
+
+	capacity := len(plain) + extra
 	size := termSize + 1 + gcm.NonceSize()
 	out := make([]byte, size, capacity)
 
@@ -949,8 +1034,20 @@ func (b *AESGCMBarrier) encrypt(path string, term uint32, gcm cipher.AEAD, plain
 	return out, nil
 }
 
+func termLabel(term uint32) []metrics.Label {
+	return []metrics.Label{
+		{
+			Name:  "term",
+			Value: strconv.FormatUint(uint64(term), 10),
+		},
+	}
+}
+
 // decrypt is used to decrypt a value using the keyring
 func (b *AESGCMBarrier) decrypt(path string, gcm cipher.AEAD, cipher []byte) ([]byte, error) {
+	if len(cipher) < 5+gcm.NonceSize() {
+		return nil, fmt.Errorf("invalid cipher length")
+	}
 	// Capture the parts
 	nonce := cipher[5 : 5+gcm.NonceSize()]
 	raw := cipher[5+gcm.NonceSize():]
@@ -986,22 +1083,32 @@ func (b *AESGCMBarrier) Encrypt(ctx context.Context, key string, plaintext []byt
 		return nil, err
 	}
 
-	ciphertext, err := b.encrypt(key, term, primary, plaintext)
+	ciphertext, err := b.encryptTracked(key, term, primary, plaintext)
 	if err != nil {
 		return nil, err
 	}
+
 	return ciphertext, nil
 }
 
 // Decrypt is used to decrypt in-memory for the BarrierEncryptor interface
-func (b *AESGCMBarrier) Decrypt(ctx context.Context, key string, ciphertext []byte) ([]byte, error) {
+func (b *AESGCMBarrier) Decrypt(_ context.Context, key string, ciphertext []byte) ([]byte, error) {
 	b.l.RLock()
 	if b.sealed {
 		b.l.RUnlock()
 		return nil, ErrBarrierSealed
 	}
 
+	if len(ciphertext) == 0 {
+		b.l.RUnlock()
+		return nil, fmt.Errorf("empty ciphertext")
+	}
+
 	// Verify the term
+	if len(ciphertext) < 4 {
+		b.l.RUnlock()
+		return nil, fmt.Errorf("invalid ciphertext term")
+	}
 	term := binary.BigEndian.Uint32(ciphertext[:4])
 
 	// Get the GCM by term
@@ -1019,7 +1126,7 @@ func (b *AESGCMBarrier) Decrypt(ctx context.Context, key string, ciphertext []by
 	// Decrypt the ciphertext
 	plain, err := b.decrypt(key, gcm, ciphertext)
 	if err != nil {
-		return nil, errwrap.Wrapf("decryption failed: {{err}}", err)
+		return nil, fmt.Errorf("decryption failed: %w", err)
 	}
 
 	return plain, nil
@@ -1034,3 +1141,128 @@ func (b *AESGCMBarrier) Keyring() (*Keyring, error) {
 
 	return b.keyring.Clone(), nil
 }
+
+func (b *AESGCMBarrier) ConsumeEncryptionCount(consumer func(int64) error) error {
+	if b.keyring != nil {
+		// Lock to prevent replacement of the key while we consume the encryptions
+		b.l.RLock()
+		defer b.l.RUnlock()
+
+		c := b.UnaccountedEncryptions.Load()
+		err := consumer(c)
+		if err == nil && c > 0 {
+			// Consumer succeeded, remove those from local encryptions
+			b.UnaccountedEncryptions.Sub(c)
+		}
+		return err
+	}
+	return nil
+}
+
+func (b *AESGCMBarrier) AddRemoteEncryptions(encryptions int64) {
+	// For rollup and persistence
+	b.UnaccountedEncryptions.Add(encryptions)
+	// For testing
+	b.RemoteEncryptions.Add(encryptions)
+}
+
+func (b *AESGCMBarrier) encryptTracked(path string, term uint32, gcm cipher.AEAD, buf []byte) ([]byte, error) {
+	ct, err := b.encrypt(path, term, gcm, buf)
+	if err != nil {
+		return nil, err
+	}
+	// Increment the local encryption count, and track metrics
+	b.UnaccountedEncryptions.Add(1)
+	b.totalLocalEncryptions.Add(1)
+	metrics.IncrCounterWithLabels(barrierEncryptsMetric, 1, termLabel(term))
+
+	return ct, nil
+}
+
+// UnaccountedEncryptions returns the number of encryptions made on the local instance only for the current key term
+func (b *AESGCMBarrier) TotalLocalEncryptions() int64 {
+	return b.totalLocalEncryptions.Load()
+}
+
+func (b *AESGCMBarrier) CheckBarrierAutoRotate(ctx context.Context) (string, error) {
+	const oneYear = 24 * 365 * time.Hour
+	reason, err := func() (string, error) {
+		b.l.RLock()
+		defer b.l.RUnlock()
+		if b.keyring != nil {
+			// Rotation Checks
+			var reason string
+
+			rc, err := b.RotationConfig()
+			if err != nil {
+				return "", err
+			}
+
+			if !rc.Disabled {
+				activeKey := b.keyring.ActiveKey()
+				ops := b.encryptions()
+				switch {
+				case activeKey.Encryptions == 0 && !activeKey.InstallTime.IsZero() && time.Since(activeKey.InstallTime) > oneYear:
+					reason = legacyRotateReason
+				case ops > rc.MaxOperations:
+					reason = "reached max operations"
+				case rc.Interval > 0 && time.Since(activeKey.InstallTime) > rc.Interval:
+					reason = "rotation interval reached"
+				}
+			}
+			return reason, nil
+		}
+		return "", nil
+	}()
+	if err != nil {
+		return "", err
+	}
+	if reason != "" {
+		return reason, nil
+	}
+
+	b.l.Lock()
+	defer b.l.Unlock()
+	if b.keyring != nil {
+		err := b.persistEncryptions(ctx)
+		if err != nil {
+			return "", err
+		}
+	}
+	return reason, nil
+}
+
+// Must be called with lock held
+func (b *AESGCMBarrier) persistEncryptions(ctx context.Context) error {
+	if !b.sealed {
+		// Encryption count persistence
+		upe := b.UnaccountedEncryptions.Load()
+		if upe > 0 {
+			activeKey := b.keyring.ActiveKey()
+			// Move local (unpersisted) encryptions to the key and persist.  This prevents us from needing to persist if
+			// there has been no activity. Since persistence performs an encryption, perversely we zero out after
+			// persistence and add 1 to the count to avoid this operation guaranteeing we need another
+			// autoRotateCheckInterval later.
+			newEncs := upe + 1
+			activeKey.Encryptions += uint64(newEncs)
+			newKeyring := b.keyring.Clone()
+			err := b.persistKeyringBestEffort(ctx, newKeyring)
+			if err != nil {
+				return err
+			}
+			b.UnaccountedEncryptions.Sub(newEncs)
+		}
+	}
+	return nil
+}
+
+// Mostly for testing, returns the total number of encryption operations performed on the active term
+func (b *AESGCMBarrier) encryptions() int64 {
+	if b.keyring != nil {
+		activeKey := b.keyring.ActiveKey()
+		if activeKey != nil {
+			return b.UnaccountedEncryptions.Load() + int64(activeKey.Encryptions)
+		}
+	}
+	return 0
+}
