@@ -1,9 +1,13 @@
-// +build !enterprise
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
+//go:build !enterprise
 
 package vault
 
 import (
-	"github.com/hashicorp/errwrap"
+	"fmt"
+
 	"github.com/hashicorp/vault/helper/namespace"
 	"github.com/hashicorp/vault/sdk/logical"
 )
@@ -21,7 +25,7 @@ func (m *ExpirationManager) collectLeases() (map[*namespace.Namespace][]string,
 	existing := make(map[*namespace.Namespace][]string)
 	keys, err := logical.CollectKeys(m.quitContext, m.leaseView(namespace.RootNamespace))
 	if err != nil {
-		return nil, 0, errwrap.Wrapf("failed to scan for leases: {{err}}", err)
+		return nil, 0, fmt.Errorf("failed to scan for leases: %w", err)
 	}
 	existing[namespace.RootNamespace] = keys
 	leaseCount += len(keys)
