@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package vault
 
 import (
@@ -5,22 +8,22 @@ import (
 	"encoding/json"
 	"reflect"
 	"strings"
-	"sync"
 	"testing"
 	"time"
 
-	metrics "github.com/armon/go-metrics"
+	"github.com/armon/go-metrics"
 	"github.com/go-test/deep"
-	"github.com/hashicorp/vault/audit"
 	"github.com/hashicorp/vault/helper/metricsutil"
 	"github.com/hashicorp/vault/helper/namespace"
+	"github.com/hashicorp/vault/helper/testhelpers/corehelpers"
+	"github.com/hashicorp/vault/helper/versions"
 	"github.com/hashicorp/vault/sdk/helper/compressutil"
 	"github.com/hashicorp/vault/sdk/helper/jsonutil"
 	"github.com/hashicorp/vault/sdk/logical"
 )
 
 func TestMount_ReadOnlyViewDuringMount(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, _ := TestCoreUnsealed(t)
 	c.logicalBackends["noop"] = func(ctx context.Context, config *logical.BackendConfig) (logical.Backend, error) {
 		err := config.StorageView.Put(ctx, &logical.StorageEntry{
 			Key:   "bar",
@@ -110,7 +113,7 @@ func TestLogicalMountMetrics(t *testing.T) {
 }
 
 func TestCore_DefaultMountTable(t *testing.T) {
-	c, keys, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, keys, _ := TestCoreUnsealed(t)
 	verifyDefaultTable(t, c.mounts, 4)
 
 	// Start a second core with same physical
@@ -118,7 +121,7 @@ func TestCore_DefaultMountTable(t *testing.T) {
 	conf := &CoreConfig{
 		Physical:        c.physical,
 		DisableMlock:    true,
-		BuiltinRegistry: NewMockBuiltinRegistry(),
+		BuiltinRegistry: corehelpers.NewMockBuiltinRegistry(),
 		MetricSink:      metricsutil.NewClusterMetricSink("test-cluster", inmemSink),
 		MetricsHelper:   metricsutil.NewMetricsHelper(inmemSink, false),
 	}
@@ -126,6 +129,7 @@ func TestCore_DefaultMountTable(t *testing.T) {
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
+	defer c2.Shutdown()
 	for i, key := range keys {
 		unseal, err := TestCoreUnseal(c2, key)
 		if err != nil {
@@ -142,7 +146,7 @@ func TestCore_DefaultMountTable(t *testing.T) {
 }
 
 func TestCore_Mount(t *testing.T) {
-	c, keys, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, keys, _ := TestCoreUnsealed(t)
 	me := &MountEntry{
 		Table: mountTableType,
 		Path:  "foo",
@@ -162,7 +166,7 @@ func TestCore_Mount(t *testing.T) {
 	conf := &CoreConfig{
 		Physical:        c.physical,
 		DisableMlock:    true,
-		BuiltinRegistry: NewMockBuiltinRegistry(),
+		BuiltinRegistry: corehelpers.NewMockBuiltinRegistry(),
 		MetricSink:      metricsutil.NewClusterMetricSink("test-cluster", inmemSink),
 		MetricsHelper:   metricsutil.NewMetricsHelper(inmemSink, false),
 	}
@@ -170,6 +174,79 @@ func TestCore_Mount(t *testing.T) {
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
+	defer c2.Shutdown()
+	for i, key := range keys {
+		unseal, err := TestCoreUnseal(c2, key)
+		if err != nil {
+			t.Fatalf("err: %v", err)
+		}
+		if i+1 == len(keys) && !unseal {
+			t.Fatalf("should be unsealed")
+		}
+	}
+
+	// Verify matching mount tables
+	if diff := deep.Equal(c.mounts.sortEntriesByPath(), c2.mounts.sortEntriesByPath()); len(diff) > 0 {
+		t.Fatalf("mismatch: %v", diff)
+	}
+}
+
+func TestCore_Mount_secrets_builtin_RunningVersion(t *testing.T) {
+	c, _, _ := TestCoreUnsealed(t)
+	me := &MountEntry{
+		Table: mountTableType,
+		Path:  "foo",
+		Type:  "generic",
+	}
+	err := c.mount(namespace.RootContext(nil), me)
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	match := c.router.MatchingMount(namespace.RootContext(nil), "foo/bar")
+	if match != "foo/" {
+		t.Fatalf("missing mount")
+	}
+
+	raw, _ := c.router.root.Get(match)
+	// we override the running version of builtins
+	if !versions.IsBuiltinVersion(raw.(*routeEntry).mountEntry.RunningVersion) {
+		t.Errorf("Expected mount to have builtin version but got %s", raw.(*routeEntry).mountEntry.RunningVersion)
+	}
+}
+
+// TestCore_Mount_kv_generic tests that we can successfully mount kv using the
+// kv alias "generic"
+func TestCore_Mount_kv_generic(t *testing.T) {
+	c, keys, _ := TestCoreUnsealed(t)
+	me := &MountEntry{
+		Table: mountTableType,
+		Path:  "foo",
+		Type:  "generic",
+	}
+	err := c.mount(namespace.RootContext(nil), me)
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	match := c.router.MatchingMount(namespace.RootContext(nil), "foo/bar")
+	if match != "foo/" {
+		t.Fatalf("missing mount")
+	}
+
+	inmemSink := metrics.NewInmemSink(1000000*time.Hour, 2000000*time.Hour)
+	conf := &CoreConfig{
+		Physical:        c.physical,
+		DisableMlock:    true,
+		BuiltinRegistry: corehelpers.NewMockBuiltinRegistry(),
+		MetricSink:      metricsutil.NewClusterMetricSink("test-cluster", inmemSink),
+		MetricsHelper:   metricsutil.NewMetricsHelper(inmemSink, false),
+	}
+	c2, err := NewCore(conf)
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	defer c2.Shutdown()
 	for i, key := range keys {
 		unseal, err := TestCoreUnseal(c2, key)
 		if err != nil {
@@ -190,12 +267,12 @@ func TestCore_Mount(t *testing.T) {
 // entries, and that upon reading the entries from both are recombined
 // correctly
 func TestCore_Mount_Local(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, _ := TestCoreUnsealed(t)
 
 	c.mounts = &MountTable{
 		Type: mountTableType,
 		Entries: []*MountEntry{
-			&MountEntry{
+			{
 				Table:            mountTableType,
 				Path:             "noop/",
 				Type:             "kv",
@@ -205,7 +282,7 @@ func TestCore_Mount_Local(t *testing.T) {
 				NamespaceID:      namespace.RootNamespaceID,
 				namespace:        namespace.RootNamespace,
 			},
-			&MountEntry{
+			{
 				Table:            mountTableType,
 				Path:             "noop2/",
 				Type:             "kv",
@@ -288,8 +365,79 @@ func TestCore_Mount_Local(t *testing.T) {
 	}
 }
 
+func TestCore_FindOps(t *testing.T) {
+	c, _, _ := TestCoreUnsealed(t)
+	uuid1 := "80DA741F-3997-4179-B531-EBD7371DFA86"
+	uuid2 := "0178594D-F267-445A-89A3-5B5DFC4A4C0F"
+	path1 := "kv1"
+	path2 := "kv2"
+
+	c.mounts = &MountTable{
+		Type: mountTableType,
+		Entries: []*MountEntry{
+			{
+				Table:            mountTableType,
+				Path:             path1,
+				Type:             "kv",
+				UUID:             "abcd",
+				Accessor:         "kv-abcd",
+				BackendAwareUUID: uuid1,
+				NamespaceID:      namespace.RootNamespaceID,
+				namespace:        namespace.RootNamespace,
+			},
+			{
+				Table:            mountTableType,
+				Path:             path2,
+				Type:             "kv",
+				UUID:             "bcde",
+				Accessor:         "kv-bcde",
+				BackendAwareUUID: uuid2,
+				NamespaceID:      namespace.RootNamespaceID,
+				namespace:        namespace.RootNamespace,
+			},
+		},
+	}
+
+	// Both should set up successfully
+	if err := c.setupMounts(namespace.RootContext(nil)); err != nil {
+		t.Fatal(err)
+	}
+
+	if len(c.mounts.Entries) != 2 {
+		t.Fatalf("expected two entries, got %d", len(c.mounts.Entries))
+	}
+
+	// Unknown uuids/paths should return nil, nil
+	entry, err := c.mounts.findByBackendUUID(namespace.RootContext(nil), "unknown")
+	if err != nil || entry != nil {
+		t.Fatalf("expected no errors nor matches got, error: %#v entry: %#v", err, entry)
+	}
+	entry, err = c.mounts.find(namespace.RootContext(nil), "unknown")
+	if err != nil || entry != nil {
+		t.Fatalf("expected no errors nor matches got, error: %#v entry: %#v", err, entry)
+	}
+
+	// Find our entry by its uuid
+	entry, err = c.mounts.findByBackendUUID(namespace.RootContext(nil), uuid1)
+	if err != nil || entry == nil {
+		t.Fatalf("failed finding entry by uuid error: %#v entry: %#v", err, entry)
+	}
+	if entry.Path != path1 {
+		t.Fatalf("found incorrect entry by uuid, entry should had a path of '%s': %#v", path1, entry)
+	}
+
+	// Find another entry by its path
+	entry, err = c.mounts.find(namespace.RootContext(nil), path2)
+	if err != nil || entry == nil {
+		t.Fatalf("failed finding entry by path error: %#v entry: %#v", err, entry)
+	}
+	if entry.BackendAwareUUID != uuid2 {
+		t.Fatalf("found incorrect entry by path, entry should had a uuid of '%s': %#v", uuid2, entry)
+	}
+}
+
 func TestCore_Unmount(t *testing.T) {
-	c, keys, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, keys, _ := TestCoreUnsealed(t)
 	err := c.unmount(namespace.RootContext(nil), "secret")
 	if err != nil {
 		t.Fatalf("err: %v", err)
@@ -304,7 +452,7 @@ func TestCore_Unmount(t *testing.T) {
 	conf := &CoreConfig{
 		Physical:        c.physical,
 		DisableMlock:    true,
-		BuiltinRegistry: NewMockBuiltinRegistry(),
+		BuiltinRegistry: corehelpers.NewMockBuiltinRegistry(),
 		MetricSink:      metricsutil.NewClusterMetricSink("test-cluster", inmemSink),
 		MetricsHelper:   metricsutil.NewMetricsHelper(inmemSink, false),
 	}
@@ -312,6 +460,7 @@ func TestCore_Unmount(t *testing.T) {
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
+	defer c2.Shutdown()
 	for i, key := range keys {
 		unseal, err := TestCoreUnseal(c2, key)
 		if err != nil {
@@ -335,7 +484,7 @@ func TestCore_Unmount_Cleanup(t *testing.T) {
 
 func testCore_Unmount_Cleanup(t *testing.T, causeFailure bool) {
 	noop := &NoopBackend{}
-	c, _, root, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, root := TestCoreUnsealed(t)
 	c.logicalBackends["noop"] = func(context.Context, *logical.BackendConfig) (logical.Backend, error) {
 		return noop, nil
 	}
@@ -441,83 +590,10 @@ func testCore_Unmount_Cleanup(t *testing.T, causeFailure bool) {
 		}
 	}
 }
-func TestCore_RemountConcurrent(t *testing.T) {
-	c2, _, _ := TestCoreUnsealed(t)
-	noop := &NoopBackend{}
-	c2.logicalBackends["noop"] = func(context.Context, *logical.BackendConfig) (logical.Backend, error) {
-		return noop, nil
-	}
 
-	// Mount the noop backend
-	mount1 := &MountEntry{
-		Table: mountTableType,
-		Path:  "test1/",
-		Type:  "noop",
-	}
-
-	if err := c2.mount(namespace.RootContext(nil), mount1); err != nil {
-		t.Fatalf("err: %v", err)
-	}
-
-	mount2 := &MountEntry{
-		Table: mountTableType,
-		Path:  "test2/",
-		Type:  "noop",
-	}
-	if err := c2.mount(namespace.RootContext(nil), mount2); err != nil {
-		t.Fatalf("err: %v", err)
-	}
-
-	mount3 := &MountEntry{
-		Table: mountTableType,
-		Path:  "test3/",
-		Type:  "noop",
-	}
-
-	if err := c2.mount(namespace.RootContext(nil), mount3); err != nil {
-		t.Fatalf("err: %v", err)
-	}
-
-	wg := &sync.WaitGroup{}
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
-		err := c2.remount(namespace.RootContext(nil), "test1", "foo", true)
-		if err != nil {
-			t.Logf("err: %v", err)
-		}
-	}()
-
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
-		err := c2.remount(namespace.RootContext(nil), "test2", "foo", true)
-		if err != nil {
-			t.Logf("err: %v", err)
-		}
-	}()
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
-		err := c2.remount(namespace.RootContext(nil), "test2", "foo", true)
-		if err != nil {
-			t.Logf("err: %v", err)
-		}
-	}()
-	wg.Wait()
-
-	c2MountMap := map[string]interface{}{}
-	for _, v := range c2.mounts.Entries {
-		
-		if _, ok := c2MountMap[v.Path]; ok {
-		t.Fatalf("duplicated mount path found at %s", v.Path)
-		}
-		c2MountMap[v.Path] = v
-	}
-}
 func TestCore_Remount(t *testing.T) {
-	c, keys, _, _ := TestCoreUnsealedWithMetrics(t)
-	err := c.remount(namespace.RootContext(nil), "secret", "foo", true)
+	c, keys, _ := TestCoreUnsealed(t)
+	err := c.remountSecretsEngineCurrentNamespace(namespace.RootContext(nil), "secret", "foo", true)
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
@@ -527,20 +603,9 @@ func TestCore_Remount(t *testing.T) {
 		t.Fatalf("failed remount")
 	}
 
-	inmemSink := metrics.NewInmemSink(1000000*time.Hour, 2000000*time.Hour)
-	conf := &CoreConfig{
-		Physical:        c.physical,
-		DisableMlock:    true,
-		BuiltinRegistry: NewMockBuiltinRegistry(),
-		MetricSink:      metricsutil.NewClusterMetricSink("test-cluster", inmemSink),
-		MetricsHelper:   metricsutil.NewMetricsHelper(inmemSink, false),
-	}
-	c2, err := NewCore(conf)
-	if err != nil {
-		t.Fatalf("err: %v", err)
-	}
+	c.sealInternal()
 	for i, key := range keys {
-		unseal, err := TestCoreUnseal(c2, key)
+		unseal, err := TestCoreUnseal(c, key)
 		if err != nil {
 			t.Fatalf("err: %v", err)
 		}
@@ -549,26 +614,15 @@ func TestCore_Remount(t *testing.T) {
 		}
 	}
 
-	// Verify matching mount tables
-	if c.mounts.Type != c2.mounts.Type {
-		t.Fatal("types don't match")
-	}
-	cMountMap := map[string]interface{}{}
-	for _, v := range c.mounts.Entries {
-		cMountMap[v.Path] = v
-	}
-	c2MountMap := map[string]interface{}{}
-	for _, v := range c2.mounts.Entries {
-		c2MountMap[v.Path] = v
-	}
-	if diff := deep.Equal(cMountMap, c2MountMap); diff != nil {
-		t.Fatal(diff)
+	match = c.router.MatchingMount(namespace.RootContext(nil), "foo/bar")
+	if match != "foo/" {
+		t.Fatalf("failed remount")
 	}
 }
 
 func TestCore_Remount_Cleanup(t *testing.T) {
 	noop := &NoopBackend{}
-	c, _, root, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, root := TestCoreUnsealed(t)
 	c.logicalBackends["noop"] = func(context.Context, *logical.BackendConfig) (logical.Backend, error) {
 		return noop, nil
 	}
@@ -624,7 +678,7 @@ func TestCore_Remount_Cleanup(t *testing.T) {
 	}
 
 	// Remount, this should cleanup
-	if err := c.remount(namespace.RootContext(nil), "test/", "new/", true); err != nil {
+	if err := c.remountSecretsEngineCurrentNamespace(namespace.RootContext(nil), "test/", "new/", true); err != nil {
 		t.Fatalf("err: %v", err)
 	}
 
@@ -652,27 +706,22 @@ func TestCore_Remount_Cleanup(t *testing.T) {
 }
 
 func TestCore_Remount_Protected(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
-	err := c.remount(namespace.RootContext(nil), "sys", "foo", true)
+	c, _, _ := TestCoreUnsealed(t)
+	err := c.remountSecretsEngineCurrentNamespace(namespace.RootContext(nil), "sys", "foo", true)
 	if err.Error() != `cannot remount "sys/"` {
 		t.Fatalf("err: %v", err)
 	}
 }
 
 func TestDefaultMountTable(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, _ := TestCoreUnsealed(t)
 	table := c.defaultMountTable()
 	verifyDefaultTable(t, table, 3)
 }
 
 func TestCore_MountTable_UpgradeToTyped(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
-
-	c.auditBackends["noop"] = func(ctx context.Context, config *audit.BackendConfig) (audit.Backend, error) {
-		return &NoopAudit{
-			Config: config,
-		}, nil
-	}
+	c, _, _ := TestCoreUnsealed(t)
+	c.auditBackends["noop"] = corehelpers.NoopAuditFactory(nil)
 
 	me := &MountEntry{
 		Table: auditTableType,
@@ -708,8 +757,8 @@ func TestCore_MountTable_UpgradeToTyped(t *testing.T) {
 func testCore_MountTable_UpgradeToTyped_Common(
 	t *testing.T,
 	c *Core,
-	testType string) {
-
+	testType string,
+) {
 	var path string
 	var mt *MountTable
 	switch testType {
@@ -795,6 +844,9 @@ func testCore_MountTable_UpgradeToTyped_Common(
 	if err != nil {
 		t.Fatal(err)
 	}
+	if entry == nil {
+		t.Fatal("nil value")
+	}
 
 	decompressedBytes, uncompressed, err := compressutil.Decompress(entry.Value)
 	if err != nil {
@@ -850,6 +902,9 @@ func verifyDefaultTable(t *testing.T, table *MountTable, expected int) {
 			if entry.Type != "system" {
 				t.Fatalf("bad: %v", entry)
 			}
+			if !entry.SealWrap {
+				t.Fatalf("expected SealWrap to be enabled: %v", entry)
+			}
 		case "identity/":
 			if entry.Type != "identity" {
 				t.Fatalf("bad: %v", entry)
@@ -868,7 +923,7 @@ func verifyDefaultTable(t *testing.T, table *MountTable, expected int) {
 }
 
 func TestSingletonMountTableFunc(t *testing.T) {
-	c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+	c, _, _ := TestCoreUnsealed(t)
 
 	mounts, auth := c.singletonMountTables()
 
@@ -899,9 +954,10 @@ func TestCore_MountInitialize(t *testing.T) {
 		backend := &InitializableBackend{
 			&NoopBackend{
 				BackendType: logical.TypeLogical,
-			}, false}
+			}, false,
+		}
 
-		c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+		c, _, _ := TestCoreUnsealed(t)
 		c.logicalBackends["initable"] = func(context.Context, *logical.BackendConfig) (logical.Backend, error) {
 			return backend, nil
 		}
@@ -924,9 +980,10 @@ func TestCore_MountInitialize(t *testing.T) {
 		backend := &InitializableBackend{
 			&NoopBackend{
 				BackendType: logical.TypeLogical,
-			}, false}
+			}, false,
+		}
 
-		c, _, _, _ := TestCoreUnsealedWithMetrics(t)
+		c, _, _ := TestCoreUnsealed(t)
 		c.logicalBackends["initable"] = func(context.Context, *logical.BackendConfig) (logical.Backend, error) {
 			return backend, nil
 		}
@@ -934,7 +991,7 @@ func TestCore_MountInitialize(t *testing.T) {
 		c.mounts = &MountTable{
 			Type: mountTableType,
 			Entries: []*MountEntry{
-				&MountEntry{
+				{
 					Table:            mountTableType,
 					Path:             "foo/",
 					Type:             "initable",
