@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package vault
 
 import (
@@ -6,13 +9,7 @@ import (
 	"github.com/hashicorp/vault/helper/namespace"
 )
 
-var (
-	NamespaceByID func(context.Context, string, *Core) (*namespace.Namespace, error) = namespaceByID
-)
-
-const (
-	mountTypeNSCubbyhole = "ns_cubbyhole"
-)
+var NamespaceByID func(context.Context, string, *Core) (*namespace.Namespace, error) = namespaceByID
 
 func namespaceByID(ctx context.Context, nsID string, c *Core) (*namespace.Namespace, error) {
 	if nsID == namespace.RootNamespaceID {
